#ifndef CELLFIELD_H
#define CELLFIELD_H

#include "Cell.h"
#include "ScalarField.h"
#include <malloc.h>

typedef struct {
	ScalarField const*	scalarField;
	Cell*				cells;
	size_t				numOfCells_cached;
} CellField;

void init_CellField( CellField* self, ScalarField const* scalarField );

static inline CellField create_CellField( ScalarField const* scalarField ) {
	CellField retval;
	init_CellField( &retval, scalarField );

	return retval;
}

static inline void destroy_CellField( CellField* self ) {
	free( self->cells );
}

static inline size_t getNumberOfCells_CellField( CellField const* self ) {
	return self->numOfCells_cached;
}

typedef struct {
	CellField*	cellField;
	Cell*		currentCell;
	uint16_t	currentIndex;
} CellFieldIterator;

static inline void init_CellFieldIterator( CellFieldIterator* self, CellField* cellField ) {
	self->cellField = cellField;
	self->currentCell = cellField->cells;
	self->currentIndex = 0;
}

static inline CellFieldIterator createIterator_CellField( CellField* self ) {
	CellFieldIterator iterator;
	init_CellFieldIterator( &iterator, self );

	return iterator;
}

static inline void destroy_CellFieldIterator( CellFieldIterator* self ) {
	if( self ) {}
}

static inline Cell* next_CellFieldIterator( CellFieldIterator* self ) {
	if( self->currentIndex < getNumberOfCells_CellField( self->cellField ) - 1 ) {
		Cell* retval = self->currentCell;

		++self->currentIndex;
		++self->currentCell;

		return retval;
	}

	return NULL;
}

#endif	/* CELLFIELD_H */
