#ifndef SCALARFIELD_H
#define SCALARFIELD_H

#include <math/vector.h>
#include <types.h>

typedef float (*potential_f)( Vector3f const* coords, void const* arg );

typedef struct {
	float		potential;
	Vector3f	position;
} ScalarFieldVertex;

typedef struct {
	ScalarFieldVertex*	verts;
	size_t				width;
	size_t				height;
	size_t				depth;
} ScalarField;

void init_ScalarField( ScalarField* self, size_t width, size_t height, size_t depth );

static inline size_t getWidth_ScalarField( ScalarField const* self ) {
	return self->width;
}

static inline size_t getHeight_ScalarField( ScalarField const* self ) {
	return self->height;
}

static inline size_t getDepth_ScalarField( ScalarField const* self ) {
	return self->depth;
}

static inline ScalarFieldVertex* getAtGrid_ScalarField( ScalarField const* self, size_t const i, size_t const j, size_t const k ) {
	return &self->verts
	[
		self->width * self->height * k +
		self->width * j +
		i
	];
}

void calculate_ScalarField( ScalarField* self, potential_f potential, void const* potential_arg );

#endif	/* SCALARFIELD_H */
