#include "ScalarField.h"
#include <malloc.h>

typedef struct {
	Vector3f	delta;
	Vector3f	current;
	Vector3f	next;
	Vector3f	start;
	Vector3i	dimensions;
	Vector3i	current_i;
} Vector3fIterator;

static inline Vector3fIterator create_Vector3fIterator( Vector3f const* start, Vector3i const* dimensions ) {
	return (Vector3fIterator) {
		.delta = create_Vector3f( 1.0f / (dimensions->x - 1), 1.0f / (dimensions->y - 1), 1.0f / (dimensions->z - 1) ),
		.current = *start,
		.next = *start,
		.start = *start,
		.dimensions = *dimensions,
		.current_i = (Vector3i) { 0, 0, 0 }
	};
}

static inline void destroy_Vector3fIterator( Vector3fIterator* self ) {
	if( self ) {}
}

static inline Vector3f* next_Vector3fIterator( Vector3fIterator* self ) {
	self->current = self->next;

	self->next.x += self->delta.x;
	if( ++self->current_i.x == self->dimensions.x ) {
		self->current_i.x = 0;
		self->next.x = self->start.x;

		self->next.y += self->delta.y;
		if( ++self->current_i.y == self->dimensions.y ) {
			self->current_i.y = 0;
			self->next.y = self->start.y;

			self->next.z += self->delta.z;
			if( ++self->current_i.z == self->dimensions.z ) {
				return NULL;
			}
		}
	}

	return &self->current;
}

void init_ScalarField( ScalarField* self, size_t const width, size_t const height, size_t const depth ) {
	*self = (ScalarField) {
		.width = width,
		.height = height,
		.depth = depth,
		.verts = malloc( sizeof( *self->verts ) * width * height * depth )
	};
}

void calculate_ScalarField( ScalarField* self, potential_f potential, void const* potential_arg ) {
	ScalarFieldVertex* vertex = self->verts;

	Vector3f const start = create_Vector3f( -0.5f, -0.5f, -0.5f );
	Vector3i const dimensions = { self->width, self->height, self->depth };

	Vector3fIterator iter = create_Vector3fIterator( &start, &dimensions );
	for( Vector3f const* coords; (coords = next_Vector3fIterator( &iter ) ); ) {
		vertex->potential = potential( coords, potential_arg );
		vertex->position = *coords;
		++vertex;
	}
	destroy_Vector3fIterator( &iter );
}
