#ifndef VECTORFIELD_H
#define VECTORFIELD_H

#include <math/vector.h>
#include <types.h>

#define VECTORFIELD_COORD_MIN	-0.5f
#define VECTORFIELD_COORD_MAX	+0.5f
#define VECTORFIELD_COORD_SPAN	(VECTORFIELD_COORD_MAX - VECTORFIELD_COORD_MIN)

typedef Vector3f (*vectorfield_f)( Vector3f const* position, void* arg );

typedef struct VectorField {
	Vector3u	dimensions;
	Vector3f*	vectors;
} VectorField;

void init_VectorField( VectorField* self, uint_t width, uint_t height, uint_t depth );

static inline VectorField create_VectorField( uint_t const width, uint_t const height, uint_t const depth ) {
	VectorField retval;
	init_VectorField( &retval, width, height, depth );

	return retval;
}

void destroy_VectorField( VectorField* self );
void calculate_VectorField( VectorField* self, vectorfield_f, void* arg );
Vector3f sample_VectorField( VectorField const* self, Vector3f const* position );

static inline Vector3f getGridStep_VectorField( VectorField const* self ) {
	return (Vector3f) {
		.x = (self->dimensions.x >= 2) ? (VECTORFIELD_COORD_SPAN / (self->dimensions.x - 1)) : 0.0f,
		.y = (self->dimensions.y >= 2) ? (VECTORFIELD_COORD_SPAN / (self->dimensions.y - 1)) : 0.0f,
		.z = (self->dimensions.z >= 2) ? (VECTORFIELD_COORD_SPAN / (self->dimensions.z - 1)) : 0.0f
	};
}

static inline Vector3u getDimensions_VectorField( VectorField const* self ) {
	return self->dimensions;
}

#endif	/* VECTORFIELD_H */
