#ifndef ACTION_H
#define ACTION_H

#include <sys/time.h>
#include <types.h>

typedef void* (*init_func)( Time const* timeFromStart, Time const* timeDuration, void* arg );
typedef void (*destroy_func)( void* context, void* arg );
typedef void (*process_func)( void* context, Time const* time, void* arg );

typedef struct {
	init_func		initFunc;
	process_func	processFunc;
	destroy_func	destroyFunc;
	void*			arg;
	void*			context;
	Time			timeStart;
	Time			timeDuration;
	bool			initialized;
} Action;

static inline void init_Action( Action* self,
	init_func initFunc,
	destroy_func destroyFunc,
	process_func processFunc,
	void* farg,
	Time const timeStart,
	Time const timeDuration
) {
	*self = (Action) {
		.initFunc = initFunc,
		.destroyFunc = destroyFunc,
		.processFunc = processFunc,
		.arg = farg,
		.context = NULL,
		.timeStart = timeStart,
		.timeDuration = timeDuration,
		.initialized = false
	};
}

static inline Action create_Action(
	init_func initFunc,
	destroy_func destroyFunc,
	process_func processFunc,
	void* farg,
	Time const timeStart,
	Time const timeDuration
) {
	Action retval;
	init_Action( &retval, initFunc, destroyFunc, processFunc, farg, timeStart, timeDuration );

	return retval;
}

static inline void destroy_Action( Action* self ) {
	if( self->initialized ) {
		self->destroyFunc( self->context, self->arg );
	}
}

static inline void destroy_Action_void( void* self ) {
	destroy_Action( self );
}

#endif	/* ACTION_H */

