#ifndef ITERATOR_H
#define ITERATOR_H

#include <types.h>

typedef struct {
	size_t		el_size;
	size_t		alignment;
	bool		(*eq_func)(void const*, void const*);
	void		(*destroy_func)(void*);
} Container;

static inline Container create_Container( 
	size_t		el_size,
	size_t		alignment,
	bool		(*eq_func)(void const*, void const*),
	void		(*destroy_func)(void*)
) {
	return (Container) {
		.el_size = el_size,
		.alignment = alignment,
		.eq_func = eq_func,
		.destroy_func = destroy_func,
	};
}

typedef struct {
	enum {
		FromTo_IteratorRange,
		From_IteratorRange,	
		To_IteratorRange,
		All_IteratorRange
	} type;
	size_t from;
	size_t to;
} IteratorRange;

typedef struct Iterator {
	Container const*	container;
	void*				(*next_func)(struct Iterator*);
} Iterator;

static inline Iterator create_Iterator( Container const* container, void* (*next_func)(struct Iterator*) ) {
	return (Iterator) {
		.container = container,
		.next_func = next_func
	};
}

static inline void* next_Iterator( Iterator* self ) {
	return self->next_func( self );
}

static inline void* find_Iterator( Iterator* self, void const* element ) {
	for( void* current_element; (current_element = next_Iterator( self )); ) {
		if( NULL == self->container->eq_func ) {
			if( current_element == element ) {
				return current_element;
			}
		} else {
			if( true == self->container->eq_func( current_element, element ) ) {
				return current_element;
			}
		}
	}

	return NULL;
}

typedef struct IteratorConst {
	Container const*	container;
	void const*			(*next_func)(struct IteratorConst*);
} IteratorConst;

static inline IteratorConst create_IteratorConst( Container const* container, void const* (*next_func)(struct IteratorConst*) ) {
	return (IteratorConst) {
		.container = container,
		.next_func = next_func
	};
}

static inline void destroy_IteratorConst( IteratorConst self ) {
	(void)self;
}

static inline void const* next_IteratorConst( IteratorConst* self ) {
	return self->next_func( self );
}

static inline void const* find_IteratorConst( IteratorConst* self, void const* element ) {
	for( void const* current_element; (current_element = next_IteratorConst( self )); ) {
		if( NULL == self->container->eq_func ) {
			if( current_element == element ) {
				return current_element;
			}
		} else {
			if( true == self->container->eq_func( current_element, element ) ) {
				return current_element;
			}
		}
	}

	return NULL;
}

#endif	/* ITERATOR_H */
