#ifndef DLTRANS_H
#define DLTRANS_H

#include <containers/vec_.h>
#include <math/vector.h>
#include <stream.h>
#include <sys/dlist.h>
#include <sys/microcode.h>
#include <sys/nativedl.h>
#include <sys/surface.h>
#include <texture.h>
#include <types.h>

#define MAX_LIGHTS_DLTRANS 	8

typedef struct {
	// TODO: consider owning referenced objects
	Lux const*		lux;
	Vector3f const*	position;
} LuxWithPosition;

typedef struct {
	uint32_t					seg0_adr;
	uint32_t					vtxOfs;
	uint16_t					lightsNum;
	float 						scale;
	bool						zBufferEnabled;
	bool						twoCycleMode;
	Mtx*						currentRspProjMtx;
	ActiveTexturesMask_DList	activeTexturesMask;
	Texture const*				currentTexture;
	VertexShader const*			currentVertexShader;
	FragmentShader const*		currentFragmentShader;

	LuxWithPosition				luxes[MAX_LIGHTS_DLTRANS];

	void const*					vtxSource;
	uint8_t						vtxFormat;
	Vector4f					clearColor;
	float						clearZVal;
	Surface						surfColor;
	Surface						surfZ;

	Vec							segSrc;								// Vec(Segment const*)
	Vec							triSrc;								// Vec(Triangle const*)
	Vec							vtxBuffer;
	Vec							rspMatrices;						// Vec(Mtx)
	Vec							cpuMtxs;							// Vec(Matrix*), managed by cmd_load_matrix() / TODO: consider owning Matrix instead of a pointer
																// (multithreading)
	Vp							rspVp;
	Lights0						rspAmbientLight;
	Vec							rspLights;							// Vec(Light[MAX_])
	Vec							materials;							// Vec(Material)
	int							geometryMode;

	//
	Command_SetBuffer const*			cmd_SetBuffer;
	Command_LoadTexture const*			cmd_LoadTexture;
	Command_SetTextureScale const*		cmd_SetTextureScale;
	Command_LoadMatrix const*			cmd_LoadMatrix;
	Command_SetCombineMode const*		cmd_SetCombineMode;
	Command_SetRenderMode const*		cmd_SetRenderMode;
	Command_NormalizePerspective const*	cmd_NormalizePerspective;
	Command_SetGeometryMode const*		cmd_SetGeometryMode;
	Command_ClearGeometryMode const*	cmd_ClearGeometryMode;
	Command_SetViewport const*			cmd_SetViewport;
	Command_SetFragmentShader const*	cmd_SetFragmentShader;
} Context_DLTrans;

typedef struct {
	bool			request;
	MicrocodeId		microcodeRequested;
	Context_DLTrans	context;
	void const*		command;
} SplitRequest_DLTrans;

typedef struct {
	Context_DLTrans	context;
} DLTrans;

void init_DLTrans( DLTrans* dltrans );
void destroy_DLTrans( DLTrans* dltrans );
SplitRequest_DLTrans process_DLTrans(
	DLTrans*				dltrans,
	Stream*					stream,
	MicrocodeId const		microcode,
	DList const*			dl_in,
	NativeDL*				dl_out,
	Context_DLTrans const*	contextOptional
);

#endif	/* DLTRANS_H */
