#ifndef INTERPOL_H
#define INTERPOL_H

#include <math/vector.h>

static inline Vector2f calculate_LERP2f( Vector2f const* v0, Vector2f const* v1, float t ) {
	return (Vector2f) {
		.x = v0->x + (v1->x - v0->x)*t,
		.y = v0->y + (v1->y - v0->y)*t
	};
}

static inline Vector3f calculate_LERP3f( Vector3f const* v0, Vector3f const* v1, float t ) {
	return (Vector3f) {
		.x = v0->x + (v1->x - v0->x)*t,
		.y = v0->y + (v1->y - v0->y)*t,
		.z = v0->z + (v1->z - v0->z)*t
	};
}

static inline Vector4f calculate_LERP4f( Vector4f const* v0, Vector4f const* v1, float t ) {
	return (Vector4f) {
		.x = v0->x + (v1->x - v0->x)*t,
		.y = v0->y + (v1->y - v0->y)*t,
		.z = v0->z + (v1->z - v0->z)*t,
		.t = v0->t + (v1->t - v0->t)*t
	};
}

static inline Vector3f calculate_Bilinear3f( Vector3f const* v0, Vector3f const* v1, Vector3f const* v2, Vector3f const* v3, float s, float t ) {
	Vector3f result0 = calculate_LERP3f( v0, v1, s );
	Vector3f result1 = calculate_LERP3f( v3, v2, s );

	return calculate_LERP3f( &result0, &result1, t );
}

#endif	/* INTERPOL_H */

