#ifndef MATERIAL_H
#define MATERIAL_H

#include <math/vector.h>
#include "texture.h"

#define TEXTURE_ID_NONE	0xFFFF

typedef enum {
	Flat_ShadingType_Material,
	Smooth_ShadingType_Material
} ShadingType_Material;

typedef enum {
	Const_TextureMapping,
	Env_TextureMapping,
	// below are deprecated names - fix lwo2ultra
	TEXTUREMAPPING_CONST = Const_TextureMapping,
	TEXTUREMAPPING_ENV = Env_TextureMapping
} TextureMapping;

typedef enum {
	Decal_TextureCombine,
	Alpha_TextureCombine,
	// below are deprecated names - fix lwo2ultra
	TEXTURECOMBINE_DECAL = Decal_TextureCombine,
	TEXTURECOMBINE_ALPHA = Alpha_TextureCombine
} TextureCombine;

typedef enum {
	// TODO: fix names
	Ccw_RenderingSide,
	Cw_RenderingSide,
	Both_RenderingSide
} RenderingSide;

typedef struct {
	Vector4f				color;
	TextureMapping			textureMapping;
	TextureCombine			textureCombine;
	ShadingType_Material	shadingType;
	uint16_t				texture0Id,
							texture1Id;
	bool					isLighted;
	RenderingSide			side;
} Material;

#define MATERIAL_INIT( col, tm, tc, texture ) {\
	.color = col,\
	.textureMapping = tm,\
	.textureCombine = tc,\
	.shadingType = Smooth_ShadingType_Material,\
	.texture0Id = texture, .texture1Id = TEXTURE_ID_NONE,\
	.isLighted = true,\
	.side = Ccw_RenderingSide\
}

static inline void init_Material( Material* self,
	Vector4f const color,
	TextureCombine const textureCombine,
	uint16_t const texture0Id, uint16_t const texture1Id
) {
	*self = (Material) {
		.color = color,
		.textureMapping = Const_TextureMapping,
		.textureCombine = textureCombine,
		.shadingType = Smooth_ShadingType_Material,
		.texture0Id = texture0Id,
		.texture1Id = texture1Id,
		.isLighted = false,
		.side = Both_RenderingSide //Ccw
	};
}

static inline Material clone_Material( Material const* self ) {
	return *self;
}

static inline void destroy_Material( Material* self ) {
	if( self ) {}
}

static inline Vector4f const* getColorConst_Material( Material const* self ) {
	return &self->color;
}

static inline Vector4f* getColor_Material( Material* self ) {
	return &self->color;
}

static inline void setColor_Material( Material* self,
	float const red,
	float const green,
	float const blue,
	float const alpha
) {
	self->color = (Vector4f) {
		.x = red,
		.y = green,
		.z = blue,
		.t = alpha
	};
}

static inline TextureCombine getTextureCombine_Material( Material const* self ) {
	return self->textureCombine;
}

static inline void setTextureCombine_Material( Material* self, TextureCombine const textureCombine ) {
	self->textureCombine = textureCombine;
}

static inline TextureMapping getTextureMapping_Material( Material const* self ) {
	return self->textureMapping;
}

static inline void setTextureMapping_Material( Material* self, TextureMapping const textureMapping ) {
	self->textureMapping = textureMapping;
}

static inline ShadingType_Material getShadingType_Material( Material const* self ) {
	return self->shadingType;
}

static inline void setShadingType_Material( Material* self, ShadingType_Material const shadingType ) {
	self->shadingType = shadingType;
}

static inline RenderingSide getRenderingSide_Material( Material const* self ) {
	return self->side;
}

static inline void setRenderingSide_Material( Material* self, RenderingSide const side ) {
	self->side = side;
}

static inline uint16_t getTextureId_Material( Material const* self ) {
	return self->texture0Id;
}

static inline void setTextureId_Material( Material* self, uint16_t const textureId ) {
	self->texture0Id = textureId;
}

static inline bool isLighted_Material( Material const* self ) {
	return self->isLighted;
}

static inline void setLighting_Material( Material* self, bool const isLighted ) {
	self->isLighted = isLighted;
}

#endif	/* MATERIAL_H */
