#ifndef MATRIX_H
#define MATRIX_H

#include <ultra64.h>
#include <math/vector.h>

//typedef float	Matrix[4][4];
typedef struct {
	float m[4][4];
} Matrix;

void invert_Matrix( Matrix const* restrict mi, Matrix* restrict mtxout );

static inline void identity_Matrix( Matrix* mtx ) {
	guMtxIdentF( mtx->m );
}

static inline void rotate_Matrix( Matrix* mtx, float a, float x, float y, float z ) {
	guRotateF( mtx->m, a, x, y, z );
}

static inline void scale_Matrix( Matrix* mtx, Vector3f const* scale ) {
	guScaleF( mtx->m, scale->x, scale->y, scale->z );
}

static inline void translate_Matrix( Matrix* mtx, float x, float y, float z ) {
	guTranslateF( mtx->m, x, y, z );
}

static inline void multiply_Matrix( Matrix* out, Matrix* ma, Matrix* mb ) {
	guMtxCatF( ma->m, mb->m, out->m );
}

static inline void perspective_Matrix( Matrix* mtx, float* perspNorm, float fovy,
	float aspect, float near, float far ) {

	u16 u16_norm;
	guPerspectiveF(
		mtx->m,
		&u16_norm,
		fovy,
		aspect,
		near,
		far,
		1.0f
	);
	*perspNorm = (float)u16_norm;
}

static inline void orthographic_Matrix( Matrix* mtx, float left, float right, float bottom,
		float top, float near, float far ) {

	guOrthoF(
		mtx->m,
		left, right,
		bottom, top,
		near, far,
		1.0f
	);
}

static inline void align_Matrix( Matrix* mtx, float angle, float x, float y, float z ) {
	guAlignF( mtx->m, angle, x, y, z );
}

#define MATRIX_GET(mtx, col, row) ((mtx).m[(col)][(row)])

#endif

