#ifndef MODELER_H
#define MODELER_H

#include <containers/vec_.h>
#include <math/vector.h>
#include <mesh.h>

typedef struct { Mesh mesh; } RingMesh;

static inline RingMesh create_RingMesh( Mesh mesh ) {
	return (RingMesh) { .mesh = mesh };
}

static inline VecIterator createInnerVerticesIterator( RingMesh* self ) {
	Layer* layer = getFirstLayer_Mesh( &self->mesh );
	return createIterator_Vec( getVertices_Layer( layer ),
		(IteratorRange) {
			.type = FromTo_IteratorRange,
			.from = 0,
			.to = getVerticesNum_Layer( layer ) / 2
		}
	);
}

static inline VecIterator createOuterVerticesIterator( RingMesh* self ) {
	Layer* layer = getFirstLayer_Mesh( &self->mesh );
	return createIterator_Vec( getVertices_Layer( layer ),
		(IteratorRange) {
			.type = FromTo_IteratorRange,
			.from = getVerticesNum_Layer( layer ) / 2,
			.to = getVerticesNum_Layer( layer )
		}
	);
}

void translateMesh_Modeler( Mesh* mesh, Vector3f const* translation );
void scaleMesh_Modeler( Mesh* mesh, Vector3f const* scale );
Mesh createSegment_Modeler( Mesh const* base, Vector3f const* normal );
Mesh createTetrahedron_Modeler( void );
Mesh createCube_Modeler( bool const createPolygons, bool const createSegments );
Mesh createSphere_Modeler( size_t const xDiv, size_t const zDiv );
Mesh createPlane_Modeler( size_t const xDivision, size_t const yDivision );
Mesh createCircle_Modeler(
	float const radius,
	size_t const sidesNum,
	bool const vertsOnly,
	bool const createPolygons,
	bool const createSegments
);
RingMesh createRing_Modeler(
	size_t const sidesNum,
	float const innerRadius,
	float const outerRadius
);
Mesh createCylinder_Modeler(
	float const radius,
	size_t const sidesNum,
	size_t const segmentsNum,
	Type_Primitives_Layer const primitiveType
);
Mesh createArrow_Modeler( void );
Mesh createUnion_Modeler( Mesh const* meshA, Mesh const* meshB );
void subdivideMesh_Modeler( Mesh* self );

#endif	/* MODELER_H */
