#ifndef CAMERA_H
#define CAMERA_H

typedef enum {
	Perspective_CameraType,
	Orthographic_CameraType
} CameraType;

typedef struct {
	float	fovY;
	float	near;
	float	far;
} PerspCamera;

typedef struct {
	float	left;
	float	right;
	float	bottom;
	float	top;
	float	near;
	float	far;
} OrthoCamera;

typedef struct {
	CameraType type;
	union {
		PerspCamera persp;
		OrthoCamera ortho;
	} cam;
} Camera;

void init_Camera( Camera* self, CameraType type );
void destroy_Camera( Camera* self );

static inline CameraType getType_Camera( Camera const* self ) {
	return self->type;
}

static inline PerspCamera* getPersp_Camera( Camera* self ) {
	return &self->cam.persp;
}

static inline OrthoCamera* getOrtho_Camera( Camera* self ) {
	return &self->cam.ortho;
}

static inline void setParams_PerspCamera( PerspCamera* self, float const fovY, float const near, float const far ) {
	self->fovY = fovY;
	self->near = near;
	self->far = far;
}

static inline float getFovY_PerspCamera( PerspCamera const* self ) {
	return self->fovY;
}

static inline float getNear_PerspCamera( PerspCamera const* self ) {
	return self->near;
}

static inline float getFar_PerspCamera( PerspCamera const* self ) {
	return self->far;
}

static inline void setParams_OrthoCamera( OrthoCamera* self,
	float const left, float const right,
	float const bottom, float const top,
	float const near, float const far ) {

	self->left = left;
	self->right = right;
	self->bottom = bottom;
	self->top = top;
	self->near = near;
	self->far = far;
}

static inline float getLeft_OrthoCamera( OrthoCamera const* self ) {
	return self->left;
}

static inline float getRight_OrthoCamera( OrthoCamera const* self ) {
	return self->right;
}

static inline float getBottom_OrthoCamera( OrthoCamera const* self ) {
	return self->bottom;
}

static inline float getTop_OrthoCamera( OrthoCamera const* self ) {
	return self->top;
}

static inline float getNear_OrthoCamera( OrthoCamera const* self ) {
	return self->bottom;
}

static inline float getFar_OrthoCamera( OrthoCamera const* self ) {
	return self->far;
}

#endif	/* CAMERA_H */

