#ifndef XFORMER_H
#define XFORMER_H

#include <math/matrix.h>
#include "coordsys.h"

struct Xformer;
typedef void (*ProcessFunc_Xformer)( struct Xformer*, void* data );

typedef struct {
	ProcessFunc_Xformer	processor;
	Matrix*			resultMtx;
	Matrix			mtx;
	Coordsys		csys;
} Xformer;

static inline Matrix* getMatrix_Xformer( Xformer* self ) {
	return &self->mtx;
}

static inline void reset_Xformer( Xformer* self ) {
	identity_Matrix( getMatrix_Xformer( self ) );
}

static inline void init_Xformer( Xformer* self, ProcessFunc_Xformer processor ) {
	self->processor	= processor;
	reset_Xformer( self );
}

static inline void destroy_Xformer( Xformer* self ) {
	if( self ) {}
}

static inline void process_Xformer( Xformer* self, void* data ) {
	if( self->processor ) {
		self->processor( (struct Xformer*)self, data );
	}
}

static inline Matrix* getResultMatrix_Xformer( Xformer const* self ) {
	return self->resultMtx;
}

static inline Coordsys const* getCoordsysConst_Xformer( Xformer const* self ) {
	return &self->csys;
}

static inline Coordsys* getCoordsys_Xformer( Xformer* self ) {
	return &self->csys;
}

static inline void transformLocally_Xformer( Xformer* self, Matrix* mtx ) {
	multiply_Matrix(
		getMatrix_Xformer( self ),
		getMatrix_Xformer( self ),
		mtx
	);
}

void finish_Xformer( Xformer* xf, Xformer* parent_xf,
	Matrix* resultMtx );

#endif	/* XFORMER_H */

