#ifndef SHADER_H
#define SHADER_H

/* vertex shader */
typedef void (*vertexShaderFunc)(
	Vertex_generic* out,
	Vertex_generic const* in,
	void const* param
);

typedef struct {
	vertexShaderFunc	shaderFunction;
	void const*			parameters;
} VertexShader;

#define STATIC_create_VertexShader( ARG_shaderFunction, ARG_parameters ) \
	{ .shaderFunction = ARG_shaderFunction, .parameters = ARG_parameters }

static inline VertexShader create_VertexShader( vertexShaderFunc shaderFunction, void const* parameters ) {
	return (VertexShader) {
		.shaderFunction = shaderFunction,
		.parameters = parameters
	};
}

static inline void destroy_VertexShader( VertexShader* self ) {
	if( self ) {}
}

static inline void execute_VertexShader( VertexShader const* self, Vertex_generic* out, Vertex_generic const* in ) {
	self->shaderFunction( out, in, self->parameters );
}

/* fragment shader */
typedef struct {
	uint_t	numberOfPasses;	
	struct {
		int a, b, c, d;
		int Aa, Ab, Ac, Ad;
	} pass[2];
} FragmentShader;

#endif	/* SHADER_H */
