#ifndef STREAM_H
#define STREAM_H

#include <types.h>

typedef struct {
	void const*	data;
	size_t		size;
	uint32_t	ofs;
} Stream;

static inline void init_Stream( Stream* self, void const* data, size_t const size ) {
	*self = (Stream) {
		.data = data,
		.size = size,
		.ofs = 0
	};
}

static inline Stream create_Stream( void const* data, size_t const size ) {
	Stream retval;
	init_Stream( &retval, data, size );

	return retval;
}

static inline void destroy_Stream( Stream const* self ) {
	if( self ) {}
}

static inline void reset_Stream( Stream* self ) {
	self->ofs = 0;
}

static inline void const* getData_Stream( Stream const* self ) {
	return &((char*)self->data)[self->ofs];
}

static inline void const* extractElement_Stream( Stream* self, size_t const size ) {
	void const* ptr = getData_Stream( self );
	self->ofs += size;

	return (self->ofs <= self->size) ? ptr : NULL;
}

static inline void rewind_Stream( Stream* self, size_t const size ) {
	self->ofs -= size;
}

#endif	/* STREAM_H */

