#ifndef DLIST_H
#define DLIST_H

#include <containers/vec_.h>
#include <layer.h>
#include <material.h>
#include <math/matrix.h>
#include <math/vector.h>
#include <node/lux.h>
#include <shader.h>
#include <vertex.h>
#include <viewport.h>
#include <texture.h>
#include <rectangle.h>
#include <triangle.h>
#include <types.h>

typedef uintmax_t	name_t;	// i wonder if this always enforces proper alignment

typedef struct {
	Vec		commands;
	size_t	num_loaded_vts;
	size_t	num_loaded_mtxs;
} DList;

typedef enum {
	Triangle_PrimitiveType_DList,
	Segment_PrimitiveType_DList
} PrimitiveType_DList;

typedef enum {
	Color_BufferType_DList	= 0x01,
	Z_BufferType_DList		= 0x02,
} BufferType_DList;

typedef enum {
	Projection_MatrixType_DList,
	Modelview_MatrixType_DList,
} MatrixType_DList;

typedef enum {
	Shade_CombineMode_DList,
	Decal_CombineMode_DList,
} CombineMode_DList;

typedef enum {
	Opaque_CombineModeAlpha_DList,
	Translucent_CombineModeAlpha_DList
} CombineModeAlpha_DList;

typedef enum {
	//Z_Opaque_RenderMode_DList,
	Opaque_RenderMode_DList,
	//Z_Translucent_RenderMode_DList,
	//Z_TexturedEdge_RenderMode_DList
	Translucent_RenderMode_DList,
	TexturedEdge_RenderMode_DList
} RenderMode_DList;

typedef enum {
	Lighting_GeometryMode_DList		= G_LIGHTING,
	TextureGen_GeometryMode_DList	= G_TEXTURE_GEN,
	//ZBuffer_GeometryMode_DList	= G_ZBUFFER,				// i think this should be on by default?
	CullBack_GeometryMode_DList		= G_CULL_BACK,
	CullFront_GeometryMode_DList	= G_CULL_FRONT,
	Shade_GeometryMode_DList		= G_SHADE,
	Gouraud_GeometryMode_DList		= G_SHADING_SMOOTH,
	Fog_GeometryMode_DList			= G_FOG
} GeometryMode_DList;

typedef enum {
	bit0_ActiveTexturesMask_DList = 0x1,
	bit1_ActiveTexturesMask_DList = 0x2,
} ActiveTexturesMask_DList;

typedef enum {
	Start_CommandName_DList,
	SetVertexSource_CommandName_DList,
	SetPrimitiveSource_CommandName_DList,
	LoadVertexBuffer_CommandName_DList,
	LoadIndexBuffer_CommandName_DList,
	DrawPrimitives_CommandName_DList,
	SetBuffer_CommandName_DList,
	ClearBuffers_CommandName_DList,
	SetClearColor_CommandName_DList,
	SetClearZValue_CommandName_DList,
	LoadTexture_CommandName_DList,
	Jump_CommandName_DList,
	End_CommandName_DList,
	SetTextureScale_CommandName_DList,
	LoadMatrix_CommandName_DList,
	LoadLight_CommandName_DList,
	Sync_CommandName_DList,
	JumpToTranslated_CommandName_DList,
	SetCombineMode_CommandName_DList,
	SetRenderMode_CommandName_DList,
	NormalizePerspective_CommandName_DList,
	SetVertexScale_CommandName_DList,
	SetLookAt_CommandName_DList,
	SetGeometryMode_CommandName_DList,
	ClearGeometryMode_CommandName_DList,
	PipeSync_CommandName_DList,
	SetViewport_CommandName_DList,
	SetMaterial_CommandName_DList,
	SetSegment_CommandName_DList,
	SetVertexShader_CommandName_DList,
	SetFragmentShader_CommandName_DList,
	SetActiveTexturesMask_CommandName_DList,
	PushState_CommandName_DList,
	PopState_CommandName_DList
} CommandName_DList;

typedef struct {
	name_t		name;
} Command_Start;

typedef struct {
	name_t		name;
	void const*	data;
	uint8_t		format;
} Command_SetVertexSource;

typedef struct {
	name_t				name;
	PrimitiveType_DList	type;
	void const*			data;
} Command_SetPrimitiveSource;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nvts;
} Command_LoadVertexBuffer;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nids;
} Command_LoadIndexBuffer;

typedef struct {
	name_t		name;
	uint8_t		type;
	uint32_t	num;
	uint32_t	ofs;
} Command_DrawPrimitives;

typedef struct {
	name_t		name;
	void*		buf;
	uint16_t	width;
	uint16_t	height;
	uint8_t		buftype;
	uint8_t		format;
} Command_SetBuffer;

typedef struct {
	name_t		name;
	uint8_t		types;
} Command_ClearBuffers;

typedef struct {
	name_t		name;
	float		r, g, b, a;
} Command_SetClearColor;

typedef struct {
	name_t		name;
	float		val;
} Command_SetClearZValue;

typedef struct {
	name_t			name;
	Texture const*	texture;
	Rectangle		tile;
	size_t			index;
	bool			prevent_seams;
} Command_LoadTexture;

typedef struct {
	name_t			name;
	DList const*	dl;
} Command_Jump;

typedef struct {
	name_t		name;
} Command_End;

typedef struct {
	name_t		name;
	float		s;
	float		t;
} Command_SetTextureScale;

typedef struct {
	name_t				name;
	Matrix*				mtx;		// Matrix const* would be better, but libultra gets a pointer to a mutable value. Not sure how unsafe dropping const is.
	MatrixType_DList	type;
} Command_LoadMatrix;

typedef struct {
	name_t			name;
	Lux const*		light;
	Vector3f const*	pos;
	uint16_t		idx;
} Command_LoadLight;

typedef struct {
	name_t		name;
} Command_Sync;

typedef struct {
	name_t		name;
} Command_PipeSync;

typedef struct {
	name_t		name;
	void const*	dl;
} Command_JumpToTranslated;

typedef struct {
	name_t					name;
	CombineMode_DList		mode;
	CombineModeAlpha_DList	amode;
} Command_SetCombineMode;

typedef struct {
	name_t				name;
	RenderMode_DList	mode;
} Command_SetRenderMode;

typedef struct {
	name_t		name;
	float		scale;
} Command_NormalizePerspective;

typedef struct {
	name_t		name;
	float		scale;
} Command_SetVertexScale;

typedef struct {
	name_t		name;
	Vector3f	eye;
	Vector3f	object;
	Vector3f	up;
} Command_SetLookAt;

typedef struct {
	name_t				name;
	GeometryMode_DList	gmode;
} Command_SetGeometryMode;

typedef struct {
	name_t				name;
	GeometryMode_DList	gmode;
} Command_ClearGeometryMode;

typedef struct {
	name_t			name;
	Viewport const*	vp;
} Command_SetViewport;

typedef struct {
	name_t		name;
	Material	mtl;
} Command_SetMaterial;

typedef struct {
	name_t		name;
	uint8_t		segment;
	void*		adr;
} Command_SetSegment;

typedef struct {
	name_t				name;
	VertexShader const*	shader;
} Command_SetVertexShader;

typedef struct {
	name_t					name;
	FragmentShader const*	shader;
} Command_SetFragmentShader;

typedef struct {
	name_t						name;
	ActiveTexturesMask_DList	activeTexturesMask;
} Command_SetActiveTexturesMask;

// TODO
typedef struct {
	name_t		name;
} Command_PushState;

// TODO
typedef struct {
	name_t		name;
} Command_PopState;

void init_DList( DList* self );
void destroy_DList( DList* self );
void reset_DList( DList* self );

static inline void const* getData_DList( DList const* self ) {
	return getHeadConst_Vec( &self->commands );
}

static inline size_t getSize_DList( DList const* self ) {
	return getLength_Vec( &self->commands );
}

#define FOO(dl, T)																	\
	int dummy;																		\
	for( uint_t i=0; i<sizeof(T); ++i ) { pushBack_Vec( &dl->commands, &dummy ); }	\
	T* p = (T*)((char*)getTail_Vec( &dl->commands ) - sizeof(T));

static inline void pushStart_DList( DList* self ) {
	FOO(self, Command_Start);
	p->name = Start_CommandName_DList;
}

static inline void pushSetVertexSource_DList( DList* self, void const* vts, uint8_t const format ) {
	FOO(self, Command_SetVertexSource);
	p->name = SetVertexSource_CommandName_DList;
	p->format = format;
	p->data = vts;
}

static inline void pushSetPrimitiveSource_DList( DList* self, PrimitiveType_DList const type, void const* ids ) {
	FOO(self, Command_SetPrimitiveSource);
	p->name = SetPrimitiveSource_CommandName_DList;
	p->type = type;
	p->data = ids;
}

static inline void pushLoadVertexBuffer_DList( DList* self, uint32_t const src_ofs,
	uint32_t const dst_ofs, uint32_t nvts ) {

	FOO(self, Command_LoadVertexBuffer);
	p->name = LoadVertexBuffer_CommandName_DList;
	p->src_ofs = src_ofs;
	p->dst_ofs = dst_ofs;
	p->nvts = nvts;

	self->num_loaded_vts += nvts;
}

static inline void pushLoadIndexBuffer_DList( DList* self, uint32_t const src_ofs,
	uint32_t const dst_ofs, uint32_t const nids ) {

	FOO(self, Command_LoadIndexBuffer);
	p->name = LoadIndexBuffer_CommandName_DList;
	p->src_ofs = src_ofs;
	p->dst_ofs = dst_ofs;
	p->nids = nids;
}

static inline void pushDrawPrimitives_DList( DList* self, uint8_t const type, uint32_t const num, uint32_t const ofs ) {
	FOO(self, Command_DrawPrimitives);
	p->name = DrawPrimitives_CommandName_DList;
	p->type = type;
	p->num = num;
	p->ofs = ofs;
}

static inline void pushSetBuffer_DList( DList* self, uint8_t const buftype, uint8_t const format,
	uint16_t const width, uint16_t const height, void* buf ) {

	FOO(self, Command_SetBuffer);
	p->name = SetBuffer_CommandName_DList;
	p->buftype = buftype;
	p->format = format;
	p->width = width;
	p->height = height;
	p->buf = buf;
}

static inline void pushClearBuffers_DList( DList* self, uint8_t const types ) {
	FOO(self, Command_ClearBuffers);
	p->name = ClearBuffers_CommandName_DList;
	p->types = types;
}

static inline void pushSetClearColor_DList( DList* self, float const r, float const g, float const b, float const a ) {
	FOO(self, Command_SetClearColor);
	p->name = SetClearColor_CommandName_DList;
	p->r = r;
	p->g = g;
	p->b = b;
	p->a = a;
}

static inline void pushSetClearZValue_DList( DList* self, float const val ) {
	FOO(self, Command_SetClearZValue);
	p->name = SetClearZValue_CommandName_DList;
	p->val = val;
}

static inline void pushLoadTexture_DList( DList* self, Texture const* texture, Rectangle const tile, size_t const index, bool const prevent_seams ) {
	FOO(self, Command_LoadTexture);
	p->name = LoadTexture_CommandName_DList;
	p->texture = texture;
	p->tile = tile;
	p->index = index;
	p->prevent_seams = prevent_seams;
}

static inline void pushJump_DList( DList* self, DList const* dl ) {
	FOO(self, Command_Jump);
	p->name = Jump_CommandName_DList;
	p->dl = dl;
}

static inline void pushEnd_DList( DList* self ) {
	FOO(self, Command_End);
	p->name = End_CommandName_DList;
}

static inline void pushSetTextureScale_DList( DList* self, float const s, float const t ) {
	FOO(self, Command_SetTextureScale);
	p->name = SetTextureScale_CommandName_DList;
	p->s = s;
	p->t = t;
}

static inline void pushLoadMatrix_DList( DList* self, Matrix* mtx, MatrixType_DList const type ) {
	FOO(self, Command_LoadMatrix);
	p->name = LoadMatrix_CommandName_DList;
	p->mtx = mtx;
	p->type = type;

	++self->num_loaded_mtxs;
}

static inline void pushLoadLight_DList( DList* self, Lux const* light,
	Vector3f const* position, uint16_t const idx ) {

	FOO(self, Command_LoadLight);
	p->name = LoadLight_CommandName_DList;
	p->light = light;
	p->pos = position;
	p->idx = idx;
}

static inline void pushSync_DList( DList* self ) {
	FOO(self, Command_Sync);
	p->name = Sync_CommandName_DList;
}

static inline void pushPipeSync_DList( DList* self ) {
	FOO(self, Command_PipeSync);
	p->name = PipeSync_CommandName_DList;
}

static inline void pushJumpToTranslated_DList( DList* self, void const* dl ) {
	FOO(self, Command_JumpToTranslated);
	p->name = JumpToTranslated_CommandName_DList;
	p->dl = dl;
}

static inline void pushSetCombineMode_DList( DList* self, CombineMode_DList const mode, CombineModeAlpha_DList const amode ) {
	FOO(self, Command_SetCombineMode);
	p->name = SetCombineMode_CommandName_DList;
	p->mode = mode;
	p->amode = amode;
}

static inline void pushSetRenderMode_DList( DList* self, RenderMode_DList const mode ) {
	FOO(self, Command_SetRenderMode);
	p->name = SetRenderMode_CommandName_DList;
	p->mode = mode;
}

static inline void pushNormalizePerspective_DList( DList* self, float const scale ) {
	FOO(self, Command_NormalizePerspective);
	p->name = NormalizePerspective_CommandName_DList;
	p->scale = scale;
}

static inline void pushSetVertexScale_DList( DList* self, float const scale ) {
	FOO(self, Command_SetVertexScale);
	p->name = SetVertexScale_CommandName_DList;
	p->scale = scale;
}

static inline void pushSetLookAt_DList( DList* self, Vector3f const* eye, Vector3f const* object, Vector3f const* up ) {
	FOO(self, Command_SetLookAt);
	p->name = SetLookAt_CommandName_DList;
	p->eye = *eye;
	p->object = *object;
	p->up = *up;
}

static inline void pushSetGeometryMode_DList( DList* self, GeometryMode_DList const gmode ) {
	FOO(self, Command_SetGeometryMode);
	p->name = SetGeometryMode_CommandName_DList;
	p->gmode = gmode;
}

static inline void pushClearGeometryMode_DList( DList* self, GeometryMode_DList const gmode ) {
	FOO(self, Command_ClearGeometryMode);
	p->name = ClearGeometryMode_CommandName_DList;
	p->gmode = gmode;
}

static inline void pushSetViewport_DList( DList* self, Viewport const* vp ) {
	FOO(self, Command_SetViewport);
	p->name = SetViewport_CommandName_DList;
	p->vp = vp;
}

static inline void pushSetMaterial_DList( DList* self, Material const* material ) {
	FOO(self, Command_SetMaterial);
	p->name = SetMaterial_CommandName_DList;
	p->mtl = clone_Material( material );
}

static inline void pushSetSegment_DList( DList* self, uint8_t const segment, void* adr ) {
	FOO(self, Command_SetSegment);
	p->name = SetSegment_CommandName_DList;
	p->segment = segment;
	p->adr = adr;
}

static inline void pushSetVertexShader_DList( DList* self, VertexShader const* shader ) {
	FOO(self, Command_SetVertexShader);
	p->name = SetVertexShader_CommandName_DList;
	p->shader = shader;
}

static inline void pushSetFragmentShader_DList( DList* self, FragmentShader const* shader ) {
	FOO(self, Command_SetFragmentShader);
	p->name = SetFragmentShader_CommandName_DList;
	p->shader = shader;
}

static inline void pushSetActiveTexturesMask_DList( DList* self, ActiveTexturesMask_DList const activeTexturesMask ) {
	FOO(self, Command_SetActiveTexturesMask);
	p->name = SetActiveTexturesMask_CommandName_DList;
	p->activeTexturesMask = activeTexturesMask;
}

// TODO
static inline void pushPushState_DList( DList* self ) {
	FOO(self, Command_PushState);
	p->name = PushState_CommandName_DList;
}

// TODO
static inline void pushPopState_DList( DList* self ) {
	FOO(self, Command_PopState);
	p->name = PopState_CommandName_DList;
}

#endif	/* DLIST_H */

