#ifndef SURFACE_H
#define SURFACE_H

#include <abort.h>
#include <types.h>

typedef enum {
	Rgba5551_PixelFormat,
	Int16_PixelFormat
} PixelFormat;

typedef struct {
	PixelFormat	pixelFormat;
	void*		data;
	uint16_t	width;
	uint16_t	height;
} Surface;

static inline void init_Surface( Surface* self, PixelFormat const pixelFormat, void* data,
	uint16_t const width, uint16_t const height ) {

	*self = (Surface) {
		.pixelFormat = pixelFormat,
		.data = data,
		.width = width,
		.height = height
	};
}

static inline Surface create_Surface( PixelFormat const pixelFormat, void* data,
	uint16_t const width, uint16_t const height ) {

	Surface self;
	init_Surface( &self, pixelFormat, data, width, height );

	return self;
}

static inline PixelFormat getPixelFormat_Surface( Surface const* self ) {
	return self->pixelFormat;
}

static inline void* getData_Surface( Surface const* self ) {
	return self->data;
}

static inline bool coordInBounds_Surface( Surface const* self, int16_t const x, int16_t const y ) {
	return
		(x >= 0) && (x < self->width) &&
		(y >= 0) && (y < self->height);
}

static inline void const* getValue_Surface( Surface const* self, int16_t const x, int16_t const y ) {
	if( !coordInBounds_Surface( self, x, y ) ) {
		return NULL;
	}

	switch( getPixelFormat_Surface( self ) ) {
		case Rgba5551_PixelFormat:
		case Int16_PixelFormat:
			return &((uint16_t*)self->data)[y * self->width + x];
		default: {
			abort();
		}
	}
}

static inline uint16_t getWidth_Surface( Surface const* self ) {
	return self->width;
}

static inline uint16_t getHeight_Surface( Surface const* self ) {
	return self->height;
}

static inline float getAspectRatio_Surface( Surface const* self ) {
	return (float)getWidth_Surface( self ) / getHeight_Surface( self );
}

static inline uint32_t getPixelsNum_Surface( Surface const* self ) {
	return self->width * self->height;
}

static inline uint32_t getSize_Surface( Surface const* self ) {
	switch( getPixelFormat_Surface( self ) ) {
		case Rgba5551_PixelFormat:
		case Int16_PixelFormat:
			return getPixelsNum_Surface( self ) * 2;
		default:
			abort();
	}
}

static inline void const* getLastByte_Surface( Surface const* self ) {
	return (void*)((uintptr_t)getData_Surface( self ) + getSize_Surface( self ) - 1);
}

#endif	/* SURFACE_H */

