#ifndef VIEWPORT_H
#define VIEWPORT_H

typedef struct Viewport {
	float	left;
	float	top;
	float	right;
	float	bottom;
	float	near;
	float	far;
} Viewport;

static inline void init_Viewport(
	Viewport* self,
	float const left, float const top,
	float const right, float const bottom,
	float const near, float const far
) {
	*self = (Viewport) {
		.left = left,
		.top = top,
		.right = right,
		.bottom = bottom,
		.near = near,
		.far = far
	};
}

static inline Viewport create_Viewport(
	float const left, float const top,
	float const right, float const bottom,
	float const near, float const far
) {
	Viewport retval;
	init_Viewport( &retval, left, top, right, bottom, near, far );

	return retval;
}

static inline void destroy_Viewport( Viewport* self ) {
	if( self ) {}
}

static inline Viewport clone_Viewport( Viewport const* self ) {
	return *self;
}

static inline float getTop_Viewport( Viewport const* self ) {
	return self->top;
}

static inline float getLeft_Viewport( Viewport const* self ) {
	return self->left;
}

static inline float getBottom_Viewport( Viewport const* self ) {
	return self->bottom;
}

static inline float getRight_Viewport( Viewport const* self ) {
	return self->right;
}

static inline float getWidth_Viewport( Viewport const* self ) {
	return self->right - self->left;
}

static inline float getHeight_Viewport( Viewport const* self ) {
	return self->bottom - self->top;
}

static inline float getNear_Viewport( Viewport const* self ) {
	return self->near;
}

static inline float getFar_Viewport( Viewport const* self ) {
	return self->far;
}

static inline float getAspectRatio_Viewport( Viewport const* self ) {
	return getWidth_Viewport( self ) / getHeight_Viewport( self );
}

#endif	/* VIEWPORT_H */

