#include "VectorField.h"
#include <assert.h>
#include <interpol/interpol.h>
#include <malloc.h>

void init_VectorField( VectorField* self, uint_t width, uint_t height, uint_t depth ) {
	self->dimensions = (Vector3u) { width, height, depth };
	self->vectors = malloc( width*height*depth * sizeof(*self->vectors) );
}

void destroy_VectorField( VectorField* self ) {
	free( self->vectors );
}

void calculate_VectorField( VectorField* self, vectorfield_f calculate_f, void* arg ) {
	Vector3f* vector = self->vectors;

	Vector3f position = (Vector3f) {
		.x = VECTORFIELD_COORD_MIN,
		.y = VECTORFIELD_COORD_MIN,
		.z = VECTORFIELD_COORD_MIN
	};
	Vector3f position_delta = getGridStep_VectorField( self );

	for( uint_t k = 0; k < self->dimensions.z; ++k ) {
		for( uint_t j = 0; j < self->dimensions.y; ++j ) {
			for( uint_t i = 0; i < self->dimensions.x; ++i ) {
				*vector = calculate_f( &position, arg );
				++vector;

				position.x += position_delta.x;
			}

			position.y += position_delta.y;
		}

		position.z += position_delta.z;
	}
}

Vector3f getByIndices_VectorField( VectorField const* self, uint_t x, uint_t y, uint_t z ) {
	return self->vectors[
		z * (self->dimensions.z * self->dimensions.y) +
		y * (self->dimensions.x) +
		x
	];
}

Vector3f sample_VectorField( VectorField const* self, Vector3f const* position ) {
	assert(
		isBounded_Vector3f( position,
			&(Vector3f) { VECTORFIELD_COORD_MAX, VECTORFIELD_COORD_MAX, VECTORFIELD_COORD_MAX }
		)
	);

	Vector3f tilePreciseCoords = scaleComponents_Vector3f(
		&VECTOR3F_TEMP(
			sub_Vector3f( position,
				&(Vector3f) { VECTORFIELD_COORD_MIN, VECTORFIELD_COORD_MIN, VECTORFIELD_COORD_MIN }
			)
		),
		&(Vector3f) { self->dimensions.x - 1, self->dimensions.y - 1, self->dimensions.z - 1 }
	);

	Vector3i tileTruncatedCoords = fromVector3f_Vector3i( &tilePreciseCoords );
	float s = tilePreciseCoords.x - tileTruncatedCoords.y;
	float t = tilePreciseCoords.y - tileTruncatedCoords.y;

	Vector3f sample00 = getByIndices_VectorField( self,
		tileTruncatedCoords.x,
		tileTruncatedCoords.y,
		tileTruncatedCoords.z
	);
	Vector3f sample01 = getByIndices_VectorField( self,
		tileTruncatedCoords.x + 1,
		tileTruncatedCoords.y,
		tileTruncatedCoords.z
	);
	Vector3f sample11 = getByIndices_VectorField( self,
		tileTruncatedCoords.x + 1,
		tileTruncatedCoords.y + 1,
		tileTruncatedCoords.z
	);
	Vector3f sample10 = getByIndices_VectorField( self,
		tileTruncatedCoords.x,
		tileTruncatedCoords.y + 1,
		tileTruncatedCoords.z
	);

	return calculate_Bilinear3f(
		&sample00, &sample01, &sample11, &sample10,
		s, t
	);
}
