#include <string.h>
#include <sys/region.h>
#include <containers/stack.h>

void init_Stack( Stack* self, size_t el_size, void (*destroy_func)(void*) ) {
	init_Vec( &self->data, el_size, Default_RegionAlignment, destroy_func );
	self->el_size = el_size;
	reset_Stack( self );
}

void destroy_Stack( Stack* self ) {
	destroy_Vec( &self->data );
}

void push_Stack( Stack* self, void* element ) {
	if( getLength_Vec( &self->data ) <= self->el_pushed ) {
		resize_Vec( &self->data, getLength_Vec( &self->data ) + 16 );
	}

	void* top = getElement_Vec( &self->data, self->el_pushed++ );
	memcpy( top, element, self->el_size );
}

void* pop_Stack( Stack* self ) {
	if( 0 == self->el_pushed ) {
		return NULL;
	}

	return getElement_Vec( &self->data, --self->el_pushed );
}

void* get_Stack( Stack* self, uint_t i ) {
	if( i+1 > self->el_pushed ) {
		return NULL;
	}

	return getElement_Vec( &self->data, self->el_pushed - i - 1 );
}

inline bool isEmpty_Stack( Stack const* self ) {
	return 0 == self->el_pushed;
}

inline void reset_Stack( Stack* self ) {
	self->el_pushed = 0;
}

