#include <abort.h>
#include <assert.h>
#include <containers/vec_.h>

void init_Vec( Vec* self, size_t const el_size, size_t const alignment, void (*destroy_func)(void*) ) {
	self->base = create_Container( el_size, alignment, NULL, destroy_func );

	self->num_el = 0;
	self->capacity = 16;
	self->data_unaligned =
	 	malloc( align_size_vec( el_size, self->capacity, alignment ) );
	assert( NULL != self->data_unaligned );

	self->data = ALIGN_ADR( self->data_unaligned, alignment );
	self->end = self->data;
}

void destroy_Vec( Vec* self ) {
	if( self->base.destroy_func ) {
		VecIterator elements = createIterator_Vec( self, (IteratorRange) { .type = All_IteratorRange } );
		for( void* element; (element = next_VecIterator( &elements )); ) {
			self->base.destroy_func( element );
		}

		destroy_VecIterator( &elements );
	}

	free( self->data_unaligned );
}

Vec clone_Vec( Vec const* self, void* (*clone_func)(void const*) ) {
	Vec cloned = create_Vec( self->base.el_size, self->base.alignment, self->base.destroy_func );

	VecConstIterator iter = createConstIterator_Vec( self, (IteratorRange) { .type = All_IteratorRange } );
	for( void const* element; (element = next_VecConstIterator( &iter )); ) {
		pushBack_Vec( &cloned, clone_func ? clone_func( element ) : element );
	}
	destroy_VecConstIterator( &iter );

	return cloned;
}

Vec* new_Vec( size_t const el_size, size_t const alignment, void (*destroy_func)(void*) ) {
	Vec* vec = malloc( sizeof *vec );
	init_Vec( vec, el_size, alignment, destroy_func );

	return vec;
}

void append_Vec( Vec* self, Vec const* other ) {
	VecConstIterator iteratorOther = createConstIterator_Vec( other, (IteratorRange) { .type = All_IteratorRange } );
	for( void const* element; (element = next_VecConstIterator( &iteratorOther )); ) {
		pushBack_Vec( self, element );
	}
	destroy_VecConstIterator( &iteratorOther );
}

