#include <abort.h>
#include <assert.h>
#include <dltrans.h>
#include <malloc.h>
#include <math/math.h>
#include <math/vector.h>
#include <node/lux.h>
#include <node/xformer.h>
#include <rectangle.h>
#include <shader.h>
#include <sys/gfxdevice.h>
#include <sys/region.h>
#include <sys/sys.h>
#include <types.h>
#include <ultra64.h>
#include <vertex.h>

#define JUMP_NATIVE_DL_SEGMENT	0x01
//#define IDXBUFFER_MAXELEMENTS	64
#define MAX_NODES 				256		//+256
#define	MAX_MATERIALS 			256		//+256
#define MAX_TRISOURCES			256		//+256
#define MAX_SEGSOURCES			256		//+256

static void shader0_func( Vertex_generic* out, Vertex_generic const* in, void const* param ) {
	if( param ) {}
	*out = *in;
}
static VertexShader const defaultVertexShader = STATIC_create_VertexShader( shader0_func, NULL );
static FragmentShader const defaultFragmentShader = {
	.numberOfPasses = 2,
	.pass[0] = {
		G_CCMUX_TEXEL0, G_CCMUX_0, G_CCMUX_PRIMITIVE, G_CCMUX_0,
		G_ACMUX_TEXEL0, G_ACMUX_0, G_ACMUX_PRIMITIVE, G_ACMUX_0
	},
	.pass[1] = {
		G_CCMUX_TEXEL0, G_CCMUX_0, G_CCMUX_PRIMITIVE, G_CCMUX_0,
		G_ACMUX_TEXEL0, G_ACMUX_0, G_ACMUX_PRIMITIVE, G_ACMUX_0
	},
};

//static LookAt lookat[MAX_NODES];

static Gfx const rspinit_dl[] = {
	gsSPNumLights( NUMLIGHTS_0 ),
	gsSPClipRatio( FRUSTRATIO_1 ),
	gsSPSetGeometryMode( G_ZBUFFER ),
	gsSPEndDisplayList()
};

static Gfx const rdpinit_dl[] = {
	gsDPSetCycleType( G_CYC_1CYCLE ),
	gsDPPipelineMode( G_PM_NPRIMITIVE ),
	gsDPSetScissor( G_SC_NON_INTERLACE, 0, 0, DEFAULT_BUFFER_W, DEFAULT_BUFFER_H ),
	gsDPSetTextureLOD( G_TL_TILE ),
	gsDPSetTextureLUT( G_TT_NONE ),
	gsDPSetTextureDetail( G_TD_CLAMP ),
	gsDPSetTexturePersp( G_TP_PERSP ),
	gsDPSetTextureFilter( G_TF_BILERP ),
	gsDPSetTextureConvert( G_TC_FILT ),
	gsDPSetCombineMode( G_CC_SHADE, G_CC_SHADE ),
	gsDPSetCombineKey( G_CK_NONE ),
	gsDPSetAlphaCompare( G_AC_NONE ),
	gsDPSetRenderMode( G_RM_OPA_SURF, G_RM_OPA_SURF2 ),
	gsDPSetFogColor( 255, 255, 255, 255 ),
	gsDPSetEnvColor( 255, 255, 255, 255 ),
	gsDPSetColorDither( G_CD_DISABLE ),
	///gsDPPipeSync(),
	gsSPEndDisplayList()
};

void reset_Context_DLTrans( Context_DLTrans* self ) {
	self->vtxOfs				= 0;
	self->lightsNum				= 0;
	self->scale					= 16384.0f / 128.0f;
	self->zBufferEnabled		= true;
	self->twoCycleMode			= false;
	self->currentRspProjMtx		= NULL;
	self->activeTexturesMask	= bit0_ActiveTexturesMask_DList;
	self->currentTexture		= NULL;
	self->currentVertexShader	= &defaultVertexShader;
	self->currentFragmentShader	= &defaultFragmentShader;
	self->rspAmbientLight		= (Lights0) gdSPDefLights0( 0x00, 0x00, 0x40 );
	self->rspVp = (Vp) {
		{
			//{ (BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ/2, 0 },			// G_MAXZ = 0x03FF
			//{ (BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ/2, 0 }
			{ (DEFAULT_BUFFER_W / 2) * 4, (DEFAULT_BUFFER_H / 2) * 4, G_MAXZ / 2, 0 },
			{ (DEFAULT_BUFFER_W / 2) * 4, (DEFAULT_BUFFER_H / 2) * 4, G_MAXZ / 2, 0 }
		}
	};
	self->geometryMode				= G_ZBUFFER;			// see rspinit_dl

	self->cmd_SetBuffer				= NULL;
	self->cmd_LoadTexture			= NULL;
	self->cmd_SetTextureScale		= NULL;
	self->cmd_LoadMatrix			= NULL;
	self->cmd_SetCombineMode		= NULL;
	self->cmd_SetRenderMode			= NULL;
	self->cmd_NormalizePerspective	= NULL;
	self->cmd_SetGeometryMode		= NULL;
	self->cmd_ClearGeometryMode		= NULL;
	self->cmd_SetViewport			= NULL;
	self->cmd_SetFragmentShader		= NULL;
}

void init_Context_DLTrans( Context_DLTrans* self ) {
	*self = (Context_DLTrans) {
		// fields omitted
		.segSrc = create_Vec( sizeof(Segment const*), Default_RegionAlignment, NULL ),
		.triSrc = create_Vec( sizeof(Triangle const*), Default_RegionAlignment, NULL ),
		.vtxBuffer = create_Vec( sizeof(Vtx), B8_RegionAlignment, NULL ),
		.rspMatrices = create_Vec( sizeof(Mtx), B8_RegionAlignment, NULL ),
		.cpuMtxs = create_Vec( sizeof(Matrix*), Default_RegionAlignment, NULL ),
		.rspLights = create_Vec( sizeof(Light[MAX_LIGHTS_DLTRANS]), B8_RegionAlignment, NULL ),
		.materials = create_Vec( sizeof(Material), Default_RegionAlignment, NULL )
	};
	reset_Context_DLTrans( self );

	resize_Vec( &self->segSrc, MAX_SEGSOURCES );
	resize_Vec( &self->triSrc, MAX_TRISOURCES );
	resize_Vec( &self->cpuMtxs, MAX_NODES );
	resize_Vec( &self->rspLights, MAX_LIGHTS_DLTRANS );
	resize_Vec( &self->materials, MAX_MATERIALS );
}

void destroy_Context_DLTrans( Context_DLTrans* self ) {
	destroy_Vec( &self->materials );
	destroy_Vec( &self->rspLights );
	destroy_Vec( &self->cpuMtxs );
	destroy_Vec( &self->rspMatrices );
	destroy_Vec( &self->vtxBuffer );
	destroy_Vec( &self->triSrc );
	destroy_Vec( &self->segSrc );
}

static void handle_lights_gpu(
	NativeDL*				rsp_dl,
	uint16_t const			lightsNum,
	LuxWithPosition const	luxes[static lightsNum],
	Light*					rspCurrentLights,
	Lights0*				rspAmbientLight,
	Matrix const*			objectMatrix,
	Material const*			material
) {
	for( size_t luxIdx = 0; luxIdx < lightsNum; ++luxIdx ) {
		LuxWithPosition const* currentLux = &luxes[luxIdx];
		Vector3f const center = create_Vector3f(
			MATRIX_GET(*objectMatrix, 3, 0),
			MATRIX_GET(*objectMatrix, 3, 1),
			MATRIX_GET(*objectMatrix, 3, 2)
		);
		Vector3f dir = sub_Vector3f( &center, currentLux->position );
		dir = normalize_Vector3f( &dir );

		Light_t* rspLight = &rspCurrentLights[luxIdx].l;
		rspLight->dir[0] = -127*dir.x;
		rspLight->dir[1] = -127*dir.y;
		rspLight->dir[2] = -127*dir.z;

		Vector4f const* materialColor = getColorConst_Material( material );
		rspLight->col[0] =
			rspLight->colc[0] =
				255 * currentLux->lux->color.x * materialColor->x;
		rspLight->col[1] =
			rspLight->colc[1] =
				255 * currentLux->lux->color.y * materialColor->y;
		rspLight->col[2] =
			rspLight->colc[2] =
				255 * currentLux->lux->color.z * materialColor->z;

		addCommand_NativeDL( rsp_dl,
			gsSPLight( &rspCurrentLights[luxIdx], luxIdx + 1 )
		);
	}

	addCommand_NativeDL( rsp_dl,
		gsSPLight( rspAmbientLight, lightsNum + 1 )
	);
	s32 num_lights;
	switch( lightsNum ) {
		case 0: num_lights = NUMLIGHTS_0; break;
		case 1: num_lights = NUMLIGHTS_1; break;
		case 2: num_lights = NUMLIGHTS_2; break;
		case 3: num_lights = NUMLIGHTS_3; break;
		case 4: num_lights = NUMLIGHTS_4; break;
		case 5: num_lights = NUMLIGHTS_5; break;
		case 6: num_lights = NUMLIGHTS_6; break;
		case 7: num_lights = NUMLIGHTS_7; break;
		default: num_lights = NUMLIGHTS_7;
	}
	addCommand_NativeDL( rsp_dl,
		gsSPNumLights( num_lights )
	);
}

static void cmd_start( NativeDL* rsp_dl ) {
	addCommand_NativeDL( rsp_dl,
		gsSPSegment( 0x0, 0x0 )
	);
	addCommand_NativeDL( rsp_dl,
		gsSPSegment( JUMP_NATIVE_DL_SEGMENT, 0x0 )
	);
	addCommand_NativeDL( rsp_dl,
		gsSPDisplayList( rspinit_dl )
	);
	addCommand_NativeDL( rsp_dl,
		gsSPDisplayList( rdpinit_dl )
	);
}

static void cmd_set_segment( Context_DLTrans* context, Command_SetSegment const* cmd ) {
	context->seg0_adr = (uint32_t)cmd->adr;
}

static void cmd_set_vertex_source( Context_DLTrans* context, Command_SetVertexSource const* cmd ) {
	context->vtxFormat = cmd->format;
	context->vtxSource = cmd->data;
}

static void cmd_set_primitive_source( Context_DLTrans* context, Command_SetPrimitiveSource const* cmd ) {
	Vec* primitives;
	switch( cmd->type ) {
		case Triangle_PrimitiveType_DList:	primitives = &context->triSrc; break;
		case Segment_PrimitiveType_DList:	primitives = &context->segSrc; break;
		default:							abort();
	}
	pushBack_Vec( primitives, &cmd->data );
}

static void cmd_load_vertex_buffer(
	Matrix const*					current_cpu_mtx,
	void const*						vtxSource_arg,
	Vec*							vtxBuffer,
	Mtx*							mtx_native,
	uint32_t*						vtxOfs,
	Texture const*					texture,
	Material const*					material,
	float const						scale,
	VertexShader const*				vertex_shader,
	Command_LoadVertexBuffer const*	cmd
) {
	*vtxOfs = getLength_Vec( vtxBuffer );

	Vertex_generic const*	vtxSource_ =
		(Vertex_generic const*)vtxSource_arg + cmd->src_ofs;
	Vertex_generic			vtxSourceXformed;
	Vertex_generic const*	vtxSource = &vtxSourceXformed;
	for( size_t i=0; i<cmd->nvts; ++i ) {
		execute_VertexShader( vertex_shader, &vtxSourceXformed, vtxSource_++ );

		Vtx vtx = {
			.v.ob = {
				[0] = vtxSource->pos.x * scale,
				[1] = vtxSource->pos.y * scale,
				[2] = vtxSource->pos.z * scale
			},
		};

		if( TEXTURE_ID_NONE == getTextureId_Material( material ) ) {
			vtx.v.tc[0] = 0;
			vtx.v.tc[1] = 0;
		} else {
			vtx.v.tc[0] = vtxSource->texcoords.x * (2 * 32 * (getWidth_Texture( texture ) - 1));			// -1?
			vtx.v.tc[1] = vtxSource->texcoords.y * (2 * 32 * (getHeight_Texture( texture ) - 1));
		}

		if( isLighted_Material( material ) ) {
			vtx.n.n[0] = vtxSource->normal.x * 127;
			vtx.n.n[1] = vtxSource->normal.y * 127;
			vtx.n.n[2] = vtxSource->normal.z * 127;
			vtx.n.a = 255;
		} else {
			vtx.v.cn[0] = vtxSource->color.x * 255;
			vtx.v.cn[1] = vtxSource->color.y * 255;
			vtx.v.cn[2] = vtxSource->color.z * 255;
			vtx.v.cn[3] = vtxSource->color.t * 255;
		}

		pushBack_Vec( vtxBuffer, &vtx );
	}

	Matrix mtx_scale;
	scale_Matrix( &mtx_scale,
		(Vector3f[]){ create_Vector3f(
			1.0f / scale,
			1.0f / scale,
			1.0f / scale
		)}
	);

	Matrix mtx_final;
	multiply_Matrix( &mtx_final, &mtx_scale, (Matrix*)current_cpu_mtx );	// is it safe to drop const?
	guMtxF2L( mtx_final.m, mtx_native );
}

// depends on:
// SetRenderMode										/ implicit
// LoadVertexBuffer (vtxBuffer)
// SetPrimitiveSource (triangleSources, segmentSources)
// LoadLight (lightsNum)
static SplitRequest_DLTrans cmd_draw_primitives(
	NativeDL*						rsp_dl,
	Vec const*						vtxBuffer,
	Vec const*						triangleSources,
	Vec const*						segmentSources,
	uint32_t const					vtxOfs,
	MicrocodeId						microcode,
	Command_DrawPrimitives const*	cmd
) {
	if( (Segment_PrimitiveType_DList == cmd->type) && (L3DEX2_MicrocodeId != microcode) ) {
		return (SplitRequest_DLTrans) {
			.request = true,
			.microcodeRequested = L3DEX2_MicrocodeId
		};
	} else if( (Triangle_PrimitiveType_DList == cmd->type) && (F3DEX2_MicrocodeId != microcode) ) {
		return (SplitRequest_DLTrans) {
			.request = true,
			.microcodeRequested = F3DEX2_MicrocodeId
		};
	}

	Vtx const* vtx = getConstElement_Vec( vtxBuffer, vtxOfs /*+ cmd->ofs*/ );
	switch( cmd->type ) {
		case Triangle_PrimitiveType_DList: {
			Triangle const* tri = *(Triangle const**)getLastConst_Vec( triangleSources );
			for( size_t i=0; i<cmd->num; ++i, ++tri ) {
				addCommand_NativeDL( rsp_dl,
					gsSPVertex( &vtx[ tri->ids[0] ], 1, 0 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPVertex( &vtx[ tri->ids[1] ], 1, 1 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPVertex( &vtx[ tri->ids[2] ], 1, 2 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSP1Triangle( 0, 1, 2, 0 )
				);
			}
			break;
		}
		case Segment_PrimitiveType_DList: {
			Segment const* seg = *(Segment const**)getLastConst_Vec( segmentSources );
			for( size_t i=0; i<cmd->num; ++i, ++seg ) {
				addCommand_NativeDL( rsp_dl,
					gsSPVertex( &vtx[ seg->ids[0] ], 1, 0 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPVertex( &vtx[ seg->ids[1] ], 1, 1 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSP1Triangle( 0, 1, 1, 0 )
				);
			}
			break;
		}
		default:	abort();
	}

	addCommand_NativeDL( rsp_dl,
		gsDPPipeSync()
	);

#if 0
	return (SplitRequest_DLTrans) {
		.request = true,
		.microcodeRequested = F3DEX2_MicrocodeId
	};
#else
	return (SplitRequest_DLTrans) { .request = false };
#endif
}

static void cmd_set_buffer_gpu( NativeDL* rsp_dl, Surface* surface, uint8_t const type ) {
	switch( type ) {
		case Color_BufferType_DList: {
			addCommand_NativeDL( rsp_dl,
				gsDPSetColorImage(
					G_IM_FMT_RGBA,
					G_IM_SIZ_16b,
					surface ? getWidth_Surface( surface ) : 0,
					surface ? getData_Surface( surface ) : NULL
				)
			);
			break;
		};
		case Z_BufferType_DList: {
			addCommand_NativeDL( rsp_dl,
				gsDPSetDepthImage( surface ? getData_Surface( surface ) : NULL )
			);
			break;
		}
		default:
			abort();
			break;
	}
}

static void cmd_set_buffers(
	NativeDL* rsp_dl,
	bool* zBufferEnabled,
	Surface* surfColor,
	Surface* surfZ,
	Command_SetBuffer const* cmd
) {
	Surface* surface;
	switch( cmd->buftype ) {
		case Color_BufferType_DList: {
			surface = surfColor;
			break;
		};
		case Z_BufferType_DList: {
			*zBufferEnabled = (NULL != cmd->buf);
			surface = surfZ;
			break;
		}
		default:
			abort();
			break;
	}
	init_Surface( surface, cmd->format, cmd->buf, cmd->width, cmd->height );

	cmd_set_buffer_gpu( rsp_dl, surface, cmd->buftype );
}

// TODO: save and restore gpu state?
static void clear_buffer_gpu( NativeDL* rsp_dl, Surface* surface, uint16_t const clrval ) {
	addCommand_NativeDL( rsp_dl,
		gsDPSetColorImage(
			G_IM_FMT_RGBA,
			G_IM_SIZ_16b,
			getWidth_Surface( surface ),
			getData_Surface( surface )
		)
	);
	addCommand_NativeDL( rsp_dl,
		gsDPSetCycleType( G_CYC_FILL )
	);
	addCommand_NativeDL( rsp_dl,
		gsDPSetRenderMode(
			G_RM_NOOP,
			G_RM_NOOP2
		)
	);
	addCommand_NativeDL( rsp_dl,
		gsDPSetFillColor( (clrval << 16) | clrval )
	);

	addCommand_NativeDL( rsp_dl,
		gsDPFillRectangle(
			0,
			0,
			getWidth_Surface( surface ) - 1,
			getHeight_Surface( surface ) - 1
		)
	);
	addCommand_NativeDL( rsp_dl,
		gsDPPipeSync()
	);

#if 0
	addCommand_NativeDL( rsp_dl,
		gsDPSetCycleType( G_CYC_1CYCLE )
	);
	addCommand_NativeDL( rsp_dl,
		gsDPSetRenderMode( G_RM_OPA_SURF, G_RM_OPA_SURF2 )
	);
#endif
}

static void cmd_clear_buffers( Context_DLTrans* context, NativeDL* rsp_dl, Command_ClearBuffers const* cmd ) {
	if( cmd->types & Color_BufferType_DList ) {
		clear_buffer_gpu(
			rsp_dl,
			&context->surfColor,
			GPACK_RGBA5551(
				(uint8_t)(255 * context->clearColor.x),
				(uint8_t)(255 * context->clearColor.y),
				(uint8_t)(255 * context->clearColor.z),
				1
			)
		);
	}

	if( cmd->types & Z_BufferType_DList ) {
		clear_buffer_gpu(
			rsp_dl,
			&context->surfZ,
			GPACK_ZDZ(
				(uint16_t)(G_MAXFBZ * context->clearZVal),
				0
			)
		);

		// NOTE: see notes section in Nintendo doc for gsDPSetDepthImage!!!
		// TODO
		addCommand_NativeDL( rsp_dl,
			gsDPSetColorImage(
				G_IM_FMT_RGBA,
				G_IM_SIZ_16b,
				getWidth_Surface( &context->surfColor ),
				getData_Surface( &context->surfColor )
			)
		);
	}
}

static void cmd_set_clear_color( Context_DLTrans* context, Command_SetClearColor const* cmd ) {
	context->clearColor = create_Vector4f( cmd->r, cmd->g, cmd->b, cmd->a );
}

static void cmd_set_clear_z_value( Context_DLTrans* context, Command_SetClearZValue const* cmd ) {
	context->clearZVal = cmd->val;
}

static void load_texture_rgba32( NativeDL* rsp_dl, Texture const* texture, Rectangle const* tile ) {
	addCommand_NativeDL( rsp_dl,
		gsDPLoadTextureTile(
			getPixelsConst_Texture( texture ),
			G_IM_FMT_RGBA,
			G_IM_SIZ_32b,
			getWidth_Texture( texture ),
			getHeight_Texture( texture ),
			
			tile->p0x,
			tile->p0y,
			tile->p1x,
			tile->p1y,

			0,
			G_TX_WRAP,
			G_TX_WRAP,
			log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
			log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
			G_TX_NOLOD,
			G_TX_NOLOD
		)
		/*
		gsDPLoadTextureBlock(
			texture->pixels,
			G_IM_FMT_RGBA,
			G_IM_SIZ_32b,
			texture->width,
			texture->height,
			0,
			G_TX_WRAP,
			G_TX_WRAP,
			log2uint( texture->width ),
			log2uint( texture->height ),
			G_TX_NOLOD,
			G_TX_NOLOD
		)
		*/
	);
	addCommand_NativeDL( rsp_dl, gsDPSetTextureLUT( G_TT_NONE ) );
}

static void load_texture_rgba16(
	NativeDL*			rsp_dl,
	Texture const*		texture,
	Rectangle const*	tile,
	size_t const		index,
	bool const			preventSeams
) {
	bool const s_wrap = (tile->p1x >= getWidth_Texture( texture )) ? false : false;			// true : false
	bool const t_wrap = (tile->p1y >= getHeight_Texture( texture )) ? false : false;			// true : false

	if( false == t_wrap ) {
		addCommand_NativeDL( rsp_dl,
			gsDPLoadMultiTile(
				(uint16_t const*)getPixelsConst_Texture( texture )
					- getWidth_Texture( texture ) * (true == s_wrap),
				index*256,		// tmem adr
				index,			// render tile
				G_IM_FMT_RGBA,
				G_IM_SIZ_16b,
				getWidth_Texture( texture ),
				getHeight_Texture( texture ),

				tile->p0x - (tile->p0x ? preventSeams : 0),
				tile->p0y - (tile->p0y ? preventSeams : 0),
				tile->p1x + preventSeams,
				tile->p1y + preventSeams + (true == s_wrap),

				0,
				G_TX_WRAP,
				G_TX_WRAP,
				log2_has_frac( preventSeams + 1 + (tile->p1x - tile->p0x) )
					+ log2uint( preventSeams + 1 + (tile->p1x - tile->p0x) ),
				log2_has_frac( preventSeams + 1 + (tile->p1y - tile->p0y) )
					+ log2uint( preventSeams + 1 + (tile->p1y - tile->p0y) ),
				G_TX_NOLOD,
				G_TX_NOLOD
			)
		);
	} else {	// true == t_wrap
		addCommand_NativeDL( rsp_dl,
			gsDPLoadMultiTile(
				(uint16_t const*)getPixelsConst_Texture( texture ) - getWidth_Texture( texture ) * (true == s_wrap),
				((((tile->p1x - tile->p0x + 1) * G_IM_SIZ_16b_TILE_BYTES) + 7) >> 3)
					* (texture->height - tile->p0y + (true == s_wrap)),
				0,	// rtile
				G_IM_FMT_RGBA,
				G_IM_SIZ_16b,
				getWidth_Texture( texture ),
				getHeight_Texture( texture ),

				tile->p0x,
				0,
				tile->p1x,
				tile->p1y - getHeight_Texture( texture ) + (true == s_wrap),

				0,
				G_TX_WRAP,
				G_TX_WRAP,
				log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
				log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
				G_TX_NOLOD,
				G_TX_NOLOD
			)
		);
		addCommand_NativeDL( rsp_dl,
			gsDPLoadTextureTile(
				(uint16_t const*)getPixelsConst_Texture( texture ) - getWidth_Texture( texture ) * (true == s_wrap),
				G_IM_FMT_RGBA,
				G_IM_SIZ_16b,
				getWidth_Texture( texture ),
				getHeight_Texture( texture ),

				tile->p0x,
				tile->p0y,
				tile->p1x,
				getHeight_Texture( texture ) - 1 + (true == s_wrap),

				0,
				G_TX_WRAP,
				G_TX_WRAP,
				log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
				log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
				G_TX_NOLOD,
				G_TX_NOLOD
			)
		);
		/*
		addCommand_NativeDL( rsp_dl,
			gsDPSetTileSize(
				G_TX_RENDERTILE,
				tile->p0x << G_TEXTURE_IMAGE_FRAC,
				tile->p0y << G_TEXTURE_IMAGE_FRAC,
				tile->p1x << G_TEXTURE_IMAGE_FRAC,
				tile->p1y << G_TEXTURE_IMAGE_FRAC
			)
		);
		*/
	}

	if( true == s_wrap ) {
		addCommand_NativeDL( rsp_dl,
			gsDPSetTile( G_IM_FMT_RGBA, G_IM_SIZ_16b,
				((((tile->p1x - tile->p0x + 1) * G_IM_SIZ_16b_LINE_BYTES) + 7) >> 3),	// line
				//((((tile->p1x - tile->p0x + 1) * G_IM_SIZ_16b_LINE_BYTES) + 7) >> 3),	// tmem
				0, //(((((texture->width - tile->p0x + 1) * G_IM_SIZ_16b_TILE_BYTES) + 7) >> 3)-0),	// tmem
				G_TX_RENDERTILE, 0,
				G_TX_WRAP,
				log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
				G_TX_NOLOD,
				G_TX_WRAP,
				log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
				G_TX_NOLOD
			)
		);
	}
#if 1
	addCommand_NativeDL( rsp_dl,
		gsDPSetTileSize(
			G_TX_RENDERTILE,
			(tile->p0x - preventSeams) << G_TEXTURE_IMAGE_FRAC,
			(tile->p0y - preventSeams) << G_TEXTURE_IMAGE_FRAC,
			(tile->p1x + 0) << G_TEXTURE_IMAGE_FRAC,
			(tile->p1y + 0) << G_TEXTURE_IMAGE_FRAC
		)
	);
#endif
	addCommand_NativeDL( rsp_dl,
		gsDPSetTextureLUT( G_TT_NONE )
	);
}

static void load_texture_ci4( NativeDL* rsp_dl, Texture const* texture, Rectangle const* renderTile, bool const preventSeams ) {
	if( 0 ) {
		assert( is_even_uint( getWidth_Texture( texture ) ) );
	}
	// NOTE: seems like there is a problem with textures having width or height > 512
	// (see what happens when G_TX_CLAMP)
	// (with G_TX_WRAP only tiles where coords jump from below 512 to above are affected)
	// (a behavior resembling a change of sign?)
	// UPDATE: look at documentation of gDPSetTile: there's a parameter called line and it has 9-bit precision (0-511)
	
	Rectangle const loadTile = (Rectangle) {
		//.p0x = tile->p0x - 1 * (tile->p0x == 0 ? 0 : preventSeams),
		.p0x = (renderTile->p0x - (renderTile->p0x == 0 ? 0 : preventSeams)) & ~0x1,									// should be even
		.p0y = renderTile->p0y - (renderTile->p0y == 0 ? 0 : preventSeams),
		//.p1x = tile->p1x + 1 * ((tile->p1x == getWidth_Texture( texture ) - 0) ? 0 : preventSeams),
		.p1x = (renderTile->p1x + ((renderTile->p1x == getWidth_Texture( texture ) - 1) ? 0 : preventSeams)) | 0x1,		// should be odd
		.p1y = renderTile->p1y + ((renderTile->p1y == getHeight_Texture( texture ) - 1) ? 0 : preventSeams)
	};
	assert( is_even_uint( loadTile.p0x ) );
	assert( is_odd_uint( loadTile.p1x ) );
	assert( getArea_Rectangle( &loadTile ) <= 4096 );

	addCommand_NativeDL( rsp_dl,
		gsDPLoadTextureTile_4b(
			getPixelsConst_Texture( texture ),		// timg
			G_IM_FMT_CI,							// fmt
			getWidth_Texture( texture ),			// width
			getHeight_Texture( texture ),
			
			loadTile.p0x,
			loadTile.p0y,
			loadTile.p1x,
			loadTile.p1y,

			0,
			G_TX_WRAP,
			G_TX_WRAP,
			log2uint( getWidth_Rectangle( renderTile ) ) +
				log2_has_frac( getWidth_Rectangle( renderTile ) ),
			log2uint( getHeight_Rectangle( renderTile ) ) +
				log2_has_frac( getHeight_Rectangle( renderTile ) ),
			G_TX_NOLOD,
			G_TX_NOLOD
		)
	);
#if 1
	addCommand_NativeDL( rsp_dl,
		gsDPSetTileSize( G_TX_RENDERTILE,
			renderTile->p0x << (G_TEXTURE_IMAGE_FRAC),
			renderTile->p0y << (G_TEXTURE_IMAGE_FRAC),
			renderTile->p1x << (G_TEXTURE_IMAGE_FRAC),
			renderTile->p1y << (G_TEXTURE_IMAGE_FRAC)
		)
	);
#endif
	addCommand_NativeDL( rsp_dl,
		gsDPLoadTLUT_pal16(
			0,
			getPaletteConst_Texture( texture )
		)
	);
	if( TEXTURE_FMT_CI4_IA16 == getFormat_Texture( texture ) ) {
		addCommand_NativeDL( rsp_dl,
			gsDPSetTextureLUT( G_TT_IA16 )
		);
	} else {
		addCommand_NativeDL( rsp_dl,
			gsDPSetTextureLUT( G_TT_RGBA16 )
		);
	}
}

static void cmd_load_texture_gpu( NativeDL* rsp_dl, Command_LoadTexture const* cmd ) {
	assert( 0 == ((uintptr_t)getPixelsConst_Texture( cmd->texture ) & 0x7) );

	switch( getFormat_Texture( cmd->texture ) ) {
		case TEXTURE_FMT_RGBA32:	load_texture_rgba32( rsp_dl, cmd->texture, &cmd->tile ); break;
		case TEXTURE_FMT_RGBA16:	load_texture_rgba16( rsp_dl, cmd->texture, &cmd->tile, cmd->index, cmd->prevent_seams ); break;
		case TEXTURE_FMT_CI4:
		case TEXTURE_FMT_CI4_IA16:	load_texture_ci4( rsp_dl, cmd->texture, &cmd->tile, cmd->prevent_seams );	break;
		default:					abort();
	}
}

static void cmd_load_texture( NativeDL* rsp_dl, Texture const** currentTexture, Command_LoadTexture const* cmd ) {
	*currentTexture = cmd->texture;

	cmd_load_texture_gpu( rsp_dl, cmd );
}

static void cmd_jump(
	DLTrans*				self,
	Stream*					stream,
	MicrocodeId const		microcode,
	NativeDL*				rsp_dl,
	Context_DLTrans const*	contextOptional,
	Command_Jump const*		cmd
) {
	abort();

	NativeDL* child_dl = new_NativeDL();

	process_DLTrans( self, stream, microcode, cmd->dl, child_dl, contextOptional );
	addCommand_NativeDL( rsp_dl,
		gsSPDisplayList( child_dl )
	);
}

static void cmd_end( NativeDL* rsp_dl ) {
	addCommand_NativeDL( rsp_dl,
		gsSPEndDisplayList()
	);
}

static void cmd_set_texture_scale_gpu( NativeDL* rsp_dl, Command_SetTextureScale const* cmd ) {
	addCommand_NativeDL( rsp_dl,
		gsSPTexture(
			(s32)(0x8000 * cmd->s),
			(s32)(0x8000 * cmd->t),
			0,
			G_TX_RENDERTILE,
			G_ON
		)
	);
}

static void cmd_load_matrix_gpu( NativeDL* rsp_dl, Mtx* matrix, MatrixType_DList const type ) {
	u8 param = G_MTX_LOAD | G_MTX_NOPUSH;
	switch( type ) {
		case Projection_MatrixType_DList:	param |= G_MTX_PROJECTION; break;
		case Modelview_MatrixType_DList:	param |= G_MTX_MODELVIEW; break;
		default: abort();
	}
	addCommand_NativeDL( rsp_dl,
		gsSPMatrix( matrix, param )
	);
}

static void cmd_load_matrix(
	NativeDL*					rsp_dl,
	Vec*						cpuMtxs,
	Vec*						rspMatrices,
	Mtx**						currentRspProjMtx,
	Command_LoadMatrix const*	cmd
) {
	pushBack_Vec( cpuMtxs, &cmd->mtx );
	Matrix* cpu_mtx = *(Matrix**)getLast_Vec( cpuMtxs );

	Mtx dummy;
	pushBack_Vec( rspMatrices, &dummy );
	Mtx* mtx_native = getLast_Vec( rspMatrices );

	guMtxF2L( cpu_mtx->m, mtx_native );

	if( Projection_MatrixType_DList == cmd->type ) {
		*currentRspProjMtx = mtx_native;
	}

	cmd_load_matrix_gpu( rsp_dl, mtx_native, cmd->type );
}

static void cmd_load_light( uint16_t* lightsNum, LuxWithPosition luxes[], Command_LoadLight const* cmd ) {
	//assert( cmd->idx < MAX_LIGHTS_DLTRANS );		// currently irrelevant
	assert( *lightsNum < MAX_LIGHTS_DLTRANS );

	luxes[*lightsNum].lux		= cmd->light;
	luxes[*lightsNum].position	= cmd->pos;
	++(*lightsNum);
}

static void cmd_sync( NativeDL* rsp_dl ) {
	addCommand_NativeDL( rsp_dl,
		gsDPFullSync()
	);
}

static void cmd_pipe_sync( NativeDL* rsp_dl ) {
	addCommand_NativeDL( rsp_dl,
		gsDPPipeSync()
	);
}

static SplitRequest_DLTrans cmd_jump_to_translated(
	NativeDL*						rsp_dl,
	uint32_t						seg0_adr,
	MicrocodeId						microcode,
	Command_JumpToTranslated const* cmd
) {
	if( F3DEX2_MicrocodeId != microcode ) {
		return (SplitRequest_DLTrans) {
			.request = true,
			.microcodeRequested = F3DEX2_MicrocodeId
		};
	}

	addCommand_NativeDL( rsp_dl,
		gsSPSegment( 0x0, seg0_adr )
	);
	addCommand_NativeDL( rsp_dl,
		gsSPDisplayList( (Gfx*)((uintptr_t)(cmd->dl) | (JUMP_NATIVE_DL_SEGMENT << 24)) )
	);
	addCommand_NativeDL( rsp_dl,
		gsSPSegment( 0x0, 0x0 )
	);
	addCommand_NativeDL( rsp_dl,
		gsDPPipeSync()
	);

	return (SplitRequest_DLTrans) { .request = false };
}

static void cmd_set_combine_mode_gpu( NativeDL* rsp_dl, FragmentShader const* fragmentShader ) {
	// insert pipe sync here?
	bool const secondPassIndex = (2 == fragmentShader->numberOfPasses);
	addCommand_NativeDL_( rsp_dl,
		(Gfx) {{
			_SHIFTL( G_SETCOMBINE, 24, 8 ) |
			_SHIFTL(
				GCCc0w0(
					fragmentShader->pass[0].a,
					fragmentShader->pass[0].c,
					fragmentShader->pass[0].Aa,
					fragmentShader->pass[0].Ac
				) |
				GCCc1w0(
					fragmentShader->pass[secondPassIndex].a,
					fragmentShader->pass[secondPassIndex].c
				),
				0, 24
			),
			(unsigned int)(
				GCCc0w1(
					fragmentShader->pass[0].b,
					fragmentShader->pass[0].d,
					fragmentShader->pass[0].Ab,
					fragmentShader->pass[0].Ad
				) |
				GCCc1w1(
					fragmentShader->pass[secondPassIndex].b,
					fragmentShader->pass[secondPassIndex].Aa,
					fragmentShader->pass[secondPassIndex].Ac,
					fragmentShader->pass[secondPassIndex].d,
					fragmentShader->pass[secondPassIndex].Ab,
					fragmentShader->pass[secondPassIndex].Ad
				)
			)
		}}
	);
}

// depends on:
// DrawPrimitives (primitiveType)			/ future
// PushState / PopState (zBufferEnabled)
// SetActiveTextureMask (twoCycleMode)
static void cmd_set_render_mode_gpu(
	NativeDL*						rsp_dl,
	PrimitiveType_DList const		primitiveType,
	bool const						zBufferEnabled,
	bool const						twoCycleMode,
	Command_SetRenderMode const*	cmd
) {
	u32 modeCycle1, modeCycle2;
	switch( cmd->mode ) {
		case Opaque_RenderMode_DList:
			if( Triangle_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = zBufferEnabled ? G_RM_ZB_OPA_SURF : G_RM_OPA_SURF;
				modeCycle2 = zBufferEnabled ? G_RM_ZB_OPA_SURF2 : G_RM_OPA_SURF2;
			} else if( Segment_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = zBufferEnabled ? G_RM_AA_ZB_XLU_LINE : G_RM_AA_XLU_LINE;
				modeCycle2 = zBufferEnabled ? G_RM_AA_ZB_XLU_LINE2 : G_RM_AA_XLU_LINE2;
			} else {
				abort();
			}
			break;
	#if 0
		case Opaque_RenderMode_DList:
			if( Triangle_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = G_RM_OPA_SURF;
				modeCycle2 = G_RM_OPA_SURF2;
			}
			else if( Segment_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = G_RM_AA_XLU_LINE;	//G_RM_OPA_SURF;
				modeCycle2 = G_RM_AA_XLU_LINE2;	//G_RM_OPA_SURF2;
			} else {
				abort();
			}
			break;
	#endif
		case Translucent_RenderMode_DList:
			if( Triangle_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = zBufferEnabled ? G_RM_ZB_XLU_SURF : G_RM_XLU_SURF;
				modeCycle2 = zBufferEnabled ? G_RM_ZB_XLU_SURF2 : G_RM_XLU_SURF2;
			} else if( Segment_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = zBufferEnabled ? G_RM_AA_ZB_XLU_LINE : G_RM_AA_XLU_LINE;
				modeCycle2 = zBufferEnabled ? G_RM_AA_ZB_XLU_LINE2 : G_RM_AA_XLU_LINE2;
			} else {
				abort();
			}
			break;
		case TexturedEdge_RenderMode_DList:		// TODO
			if( Triangle_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = zBufferEnabled ? G_RM_AA_ZB_TEX_EDGE : G_RM_AA_TEX_EDGE;
				modeCycle2 = zBufferEnabled ? G_RM_AA_ZB_TEX_EDGE2 : G_RM_AA_TEX_EDGE2;
			} else if( Segment_PrimitiveType_DList == primitiveType ) {
				modeCycle1 = zBufferEnabled ? G_RM_AA_ZB_TEX_EDGE : G_RM_AA_TEX_EDGE;
				modeCycle2 = zBufferEnabled ? G_RM_ZB_OPA_SURF2 : G_RM_OPA_SURF2; //G_RM_AA_ZB_TEX_EDGE2;
			} else {
				abort();
			}
			break;
		default:
			abort();
	}

	addCommand_NativeDL( rsp_dl,
		gsDPSetRenderMode(
			(true == twoCycleMode) ? G_RM_PASS : modeCycle1,
			modeCycle2
		)
	);

	// FOG
#if 0
	addCommand_NativeDL( rsp_dl,
		gsDPSetCycleType( G_CYC_2CYCLE )
	);
	addCommand_NativeDL( rsp_dl,
		gsDPSetRenderMode(
			G_RM_FOG_SHADE_A,
			G_RM_AA_ZB_TEX_EDGE2
		)
	);
	addCommand_NativeDL( rsp_dl,
		gsSPFogPosition(
			990,
			1000
		)
	);
	addCommand_NativeDL( rsp_dl,
		gsSPSetGeometryMode( G_FOG )
	);
	addCommand_NativeDL( rsp_dl,
		gsDPSetFogColor( 255, 255, 255, 255 )
	);
#endif
}

static void cmd_normalize_perspective( NativeDL* rsp_dl, Command_NormalizePerspective const* cmd ) {
	addCommand_NativeDL( rsp_dl,
		gsSPPerspNormalize( (u16) cmd->scale )
	);
}

#if 0
static void cmd_set_look_at( NativeDL* rsp_dl, uint16_t* current_lookat_i, Command_SetLookAt const* cmd ) {
	float mreflect[4][4];
	guLookAtReflectF( mreflect, &lookat[*current_lookat_i],
		cmd->eye.x, cmd->eye.y, cmd->eye.z,
		cmd->object.x, cmd->object.y, cmd->object.z,
		cmd->up.x, cmd->up.y, cmd->up.z
	);

	addCommand_NativeDL( rsp_dl,
		gsSPLookAt( &lookat[*current_lookat_i] )
	);

	++(*current_lookat_i);
}
#endif

static int cmd_set_geometry_mode_gpu( NativeDL* rsp_dl, int const geometryModeAccum, int const geometryMode ) {
	addCommand_NativeDL( rsp_dl,
		gsSPSetGeometryMode( geometryMode )
	);

	return geometryModeAccum | geometryMode;
}

static int cmd_clear_geometry_mode_gpu( NativeDL* rsp_dl, int const geometryModeAccum, int const geometryMode ) {
	addCommand_NativeDL( rsp_dl,
		gsSPClearGeometryMode( geometryMode )
	);

	return geometryModeAccum & ~(geometryMode);
}

static void cmd_set_viewport( NativeDL* rsp_dl, Vp* rspVp, Command_SetViewport const* cmd ) {
	GfxMode const* gfxMode = getCurrentMode_GfxDevice( getGfxDevice_Sys() );
	Viewport const* viewport = cmd->vp;

	rspVp->vp.vscale[0] = (gfxMode->width * getWidth_Viewport( viewport ) / 2) * 4;
	rspVp->vp.vscale[1] = (gfxMode->height * getHeight_Viewport( viewport ) / 2) * 4;
	rspVp->vp.vscale[2] = G_MAXZ * (getFar_Viewport( viewport ) - getNear_Viewport( viewport )) / 2;			//
	rspVp->vp.vscale[3] = 0;

	rspVp->vp.vtrans[0] = (gfxMode->width * (getLeft_Viewport( viewport ) + getWidth_Viewport( viewport ) / 2)) * 4;
	rspVp->vp.vtrans[1] = (gfxMode->height * (getTop_Viewport( viewport ) + getHeight_Viewport( viewport ) / 2)) * 4;
	rspVp->vp.vtrans[2] = G_MAXZ * (getFar_Viewport( viewport ) + getNear_Viewport( viewport )) / 2;			//
	rspVp->vp.vtrans[3] = 0;

	addCommand_NativeDL( rsp_dl,
		gsSPViewport( rspVp )
	);
	addCommand_NativeDL( rsp_dl,
		gsDPSetScissor(
			G_SC_NON_INTERLACE,
			gfxMode->width * getLeft_Viewport( viewport ),
			gfxMode->height * getTop_Viewport( viewport ),
			gfxMode->width * getRight_Viewport( viewport ),
			gfxMode->height * getBottom_Viewport( viewport )
		)
	);
}

static int cmd_set_material_gpu( NativeDL* rsp_dl, int const geometryModeAccum, Material const* material ) {
	Vector4f const* color = getColorConst_Material( material );
	//assert(!(color->x == 0 && color->y == 0 && color->z == 0 ));				// TODO
	//assert(!(color->t == 0 ));
	addCommand_NativeDL( rsp_dl,
		gsDPSetPrimColor(
			0, 0,
			color->x * 255,
			color->y * 255,
			color->z * 255,
			color->t * 255
		)
	);

	int culling;
	switch( getRenderingSide_Material( material ) ) {
		case Ccw_RenderingSide: { culling = G_CULL_BACK; break; }
		case Cw_RenderingSide: { culling = G_CULL_FRONT; break; }
		case Both_RenderingSide: { culling = 0; break; }
		default: { abort(); }
	}

	int geometryMode = geometryModeAccum;
	geometryMode = cmd_clear_geometry_mode_gpu( rsp_dl, geometryMode, G_CULL_BOTH );
	geometryMode = cmd_set_geometry_mode_gpu( rsp_dl, geometryMode, culling );

	return geometryMode;
	/*
	addCommand_NativeDL( rsp_dl,
		gsSPClearGeometryMode( G_CULL_BOTH )
	);
	addCommand_NativeDL( rsp_dl,
		gsSPSetGeometryMode( culling )
	);
	*/
}

static void cmd_set_material(
	NativeDL*					rsp_dl,
	Vec*						rspLights,
	Vec*						materials,
	int*						geometryMode_out,
	Command_SetMaterial const*	cmd
) {
	pushBack_Vec( materials, &cmd->mtl );
	Light dummy[MAX_LIGHTS_DLTRANS];
	pushBack_Vec( rspLights, dummy );

	*geometryMode_out = cmd_set_material_gpu( rsp_dl, *geometryMode_out, (Material const*)getLastConst_Vec( materials ) );
}

static void cmd_set_vertex_shader( VertexShader const** vertexShader, Command_SetVertexShader const* cmd ) {
	*vertexShader = cmd->shader ? cmd->shader : &defaultVertexShader;
}

static void cmd_set_fragment_shader( FragmentShader const** fragmentShader, Command_SetFragmentShader const* cmd ) {
	*fragmentShader = cmd->shader ? cmd->shader : &defaultFragmentShader;
}

static void cmd_set_active_textures_mask_gpu( NativeDL* rsp_dl, bool const twoCycleMode ) {
	addCommand_NativeDL( rsp_dl,
		gsDPSetCycleType( (true == twoCycleMode) ? G_CYC_2CYCLE : G_CYC_1CYCLE )
	);
}

static void cmd_set_active_textures_mask(
	NativeDL*								rsp_dl,
	ActiveTexturesMask_DList*				activeTexturesMask,
	bool*									twoCycleMode,
	Command_SetActiveTexturesMask const*	cmd
) {
	*activeTexturesMask = cmd->activeTexturesMask;

	if(
		(bit1_ActiveTexturesMask_DList == cmd->activeTexturesMask) ||
		(bit0_ActiveTexturesMask_DList == cmd->activeTexturesMask)
	) {
		*twoCycleMode = false;
	} else if( (bit1_ActiveTexturesMask_DList | bit0_ActiveTexturesMask_DList) == cmd->activeTexturesMask ) {
		*twoCycleMode = true;
	} else if( 0 == cmd->activeTexturesMask ) {
		*twoCycleMode = false;
	} else { abort(); }

	cmd_set_active_textures_mask_gpu( rsp_dl, *twoCycleMode );
}

void init_DLTrans( DLTrans* self ) {
	//self->tex = NULL;
	//setColor_Material( &self->material, 1.0f, 1.0f, 1.0f, 1.0f );
	init_Context_DLTrans( &self->context );
}

void destroy_DLTrans( DLTrans* self ) {
	destroy_Context_DLTrans( &self->context );
}

static void restoreContext_DLTrans( DLTrans* self, Context_DLTrans context, NativeDL* rsp_dl ) {
	self->context = context;

	cmd_start( rsp_dl );

	if( self->context.cmd_SetViewport ) {
		cmd_set_viewport( rsp_dl, &self->context.rspVp, self->context.cmd_SetViewport );
	}

	if( self->context.cmd_SetBuffer ) {
		cmd_set_buffer_gpu( rsp_dl, &self->context.surfColor, Color_BufferType_DList );
		cmd_set_buffer_gpu( rsp_dl, &self->context.surfZ, Z_BufferType_DList );
	}

	if( self->context.cmd_NormalizePerspective ) {
		cmd_normalize_perspective( rsp_dl, self->context.cmd_NormalizePerspective );
	}

	if( self->context.cmd_LoadMatrix ) {
		cmd_load_matrix_gpu( rsp_dl, self->context.currentRspProjMtx, Projection_MatrixType_DList );
		cmd_load_matrix_gpu( rsp_dl, (Mtx*)getLast_Vec( &self->context.rspMatrices ), self->context.cmd_LoadMatrix->type );
	}

	cmd_set_material_gpu( rsp_dl, 0, (Material const*)getLastConst_Vec( &self->context.materials ) );
	cmd_set_geometry_mode_gpu( rsp_dl, 0, self->context.geometryMode );

	if( self->context.cmd_LoadTexture ) {
		cmd_load_texture_gpu( rsp_dl, self->context.cmd_LoadTexture );
	}

	if( self->context.cmd_SetTextureScale ) {
		cmd_set_texture_scale_gpu( rsp_dl, self->context.cmd_SetTextureScale );
	}

	cmd_set_active_textures_mask_gpu( rsp_dl, self->context.twoCycleMode );

	// deferred into cmd_draw_primitives
	//if( self->context.cmd_SetRenderMode ) {
		//cmd_set_render_mode_gpu( rsp_dl, self->context.zBufferEnabled, self->context.twoCycleMode, self->context.cmd_SetRenderMode );
	//}

	if( self->context.cmd_SetFragmentShader ) {
		cmd_set_fragment_shader( &self->context.currentFragmentShader, self->context.cmd_SetFragmentShader );
		cmd_set_combine_mode_gpu( rsp_dl, self->context.currentFragmentShader );
	}
	//if( self->context.cmd_SetCombineMode ) {
	//	cmd_set_combine_mode_gpu( rsp_dl, self->context.currentFragmentShader );
	//}
}

SplitRequest_DLTrans process_DLTrans(
	DLTrans*				self,
	Stream*					stream,
	MicrocodeId const		microcode,
	DList const*			dl_in,
	NativeDL*				dl_out,
	Context_DLTrans const*	contextOptional
) {
	NativeDL* rsp_dl = dl_out;

	if( contextOptional ) {
		restoreContext_DLTrans( self, *contextOptional, rsp_dl );
	} else {
		reset_Context_DLTrans( &self->context );
		//TODO
		reset_Vec( &self->context.segSrc );
		reset_Vec( &self->context.triSrc );
		reset_Vec( &self->context.vtxBuffer );
		reset_Vec( &self->context.rspMatrices );
		reset_Vec( &self->context.cpuMtxs );
		reset_Vec( &self->context.rspLights );
		reset_Vec( &self->context.materials );
		resize_Vec( &self->context.vtxBuffer, dl_in->num_loaded_vts );
		resize_Vec( &self->context.rspMatrices, dl_in->num_loaded_mtxs );
	}

	name_t name;
	do {
		name = *(name_t*)getData_Stream( stream );
		switch( name ) {
			case Start_CommandName_DList: {
				(void*)extractElement_Stream( stream,
					sizeof( Command_Start )
				);

				cmd_start( rsp_dl );
				break;
			}

			case SetVertexSource_CommandName_DList: {
				Command_SetVertexSource const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetVertexSource )
					);
				cmd_set_vertex_source( &self->context, cmd );
				break;
			}

			case SetPrimitiveSource_CommandName_DList: {
				Command_SetPrimitiveSource const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetPrimitiveSource )
					);
				cmd_set_primitive_source( &self->context, cmd );
				break;
			}

			case SetVertexScale_CommandName_DList: {							// DEPRECATED
				abort();
				/*
				Command_SetVertexScale const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetVertexScale )
					);
				scale = cmd->scale;
				*/
				break;
			}

			case LoadVertexBuffer_CommandName_DList: {
				Command_LoadVertexBuffer const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_LoadVertexBuffer )
					);

				cmd_load_vertex_buffer(
					*(Matrix const**)getLastConst_Vec( &self->context.cpuMtxs ),
					self->context.vtxSource,
					&self->context.vtxBuffer,
					getLast_Vec( &self->context.rspMatrices ),
					&self->context.vtxOfs,
					self->context.currentTexture,
					(Material const*)getLastConst_Vec( &self->context.materials ),
					self->context.scale,
					self->context.currentVertexShader,
					cmd
				);
				break;
			}

			case LoadIndexBuffer_CommandName_DList: {
				abort();
				/*
				Command_LoadIndexBuffer const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_LoadIndexBuffer )
					);

				uint32_t n = cmd->nids;
				if( n > IDXBUFFER_MAXELEMENTS ) {
					abort();
				}

				uint32_t* idxSource = self->triSource->ids + cmd->src_ofs;
				uint32_t* idx = self->idxBuffer + cmd->dst_ofs;
				while( n-- ) {
					*idx++ = *idxSource++;
				}
				*/
				break;
			}

			case DrawPrimitives_CommandName_DList: {
				Command_DrawPrimitives const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_DrawPrimitives )
					);

				handle_lights_gpu(
					rsp_dl,
					self->context.lightsNum,
					self->context.luxes,
					getLast_Vec( &self->context.rspLights ),
					&self->context.rspAmbientLight,
					getLastConst_Vec( &self->context.cpuMtxs ),
					getLastConst_Vec( &self->context.materials )
				);
				if( self->context.cmd_SetRenderMode ) {
					cmd_set_render_mode_gpu(
						rsp_dl,
						cmd->type,
						self->context.zBufferEnabled,
						self->context.twoCycleMode,
						self->context.cmd_SetRenderMode
					);
				}
				SplitRequest_DLTrans request = cmd_draw_primitives(
					rsp_dl,
					&self->context.vtxBuffer,
					&self->context.triSrc,
					&self->context.segSrc,
					self->context.vtxOfs,
					microcode,
					cmd
				);
				if( true == request.request ) {
					addCommand_NativeDL( rsp_dl,
						gsDPFullSync()
					);
					addCommand_NativeDL( rsp_dl,
						gsSPEndDisplayList()
					);

					rewind_Stream( stream, sizeof *cmd );
					request.context = self->context;
					request.command = cmd;
					return request;
				}
				break;
			}

			case SetBuffer_CommandName_DList: {
				Command_SetBuffer const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetBuffer )
					);

				self->context.cmd_SetBuffer = cmd;
				cmd_set_buffers( rsp_dl, &self->context.zBufferEnabled, &self->context.surfColor, &self->context.surfZ, cmd );
				break;
			}

			case ClearBuffers_CommandName_DList: {
				Command_ClearBuffers const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_ClearBuffers )
					);

				cmd_clear_buffers( &self->context, rsp_dl, cmd );
				break;
			}

			case SetClearColor_CommandName_DList: {
				Command_SetClearColor const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetClearColor )
					);

				cmd_set_clear_color( &self->context, cmd );

				break;
			}

			case SetClearZValue_CommandName_DList: {
				Command_SetClearZValue const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetClearZValue )
					);

				cmd_set_clear_z_value( &self->context, cmd );
				break;
			}

			case LoadTexture_CommandName_DList: {
				Command_LoadTexture const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_LoadTexture )
					);

				self->context.cmd_LoadTexture = cmd;
				cmd_load_texture( rsp_dl, &self->context.currentTexture, cmd );
				break;
			}

			case Jump_CommandName_DList: {	// beware: allocation on every invocation!
				Command_Jump const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_Jump )
					);

				cmd_jump( self, stream, microcode, rsp_dl, contextOptional, cmd );
				break;
			}

			case End_CommandName_DList: {
				(void*)extractElement_Stream( stream,
					sizeof( Command_End )
				);

				cmd_end( rsp_dl );
				break;
			}

			case SetTextureScale_CommandName_DList: {
				Command_SetTextureScale const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetTextureScale )
					);

				self->context.cmd_SetTextureScale = cmd;
				cmd_set_texture_scale_gpu( rsp_dl, cmd );
				break;
			}

			case LoadMatrix_CommandName_DList: {
				Command_LoadMatrix const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_LoadMatrix )
					);

				self->context.cmd_LoadMatrix = cmd;
				cmd_load_matrix(
					rsp_dl,
					&self->context.cpuMtxs,
					&self->context.rspMatrices,
					&self->context.currentRspProjMtx,
					cmd
				);
				break;
			}

			case LoadLight_CommandName_DList: {
				Command_LoadLight const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_LoadLight )
					);

				cmd_load_light(
					&self->context.lightsNum,
					self->context.luxes,
					cmd
				);
				break;
			}

			case Sync_CommandName_DList: {
				(void*)extractElement_Stream( stream,
					sizeof( Command_Sync )
				);

				cmd_sync( rsp_dl );
				break;
			}

			case PipeSync_CommandName_DList: {
				(void*)extractElement_Stream( stream,
					sizeof( Command_PipeSync )
				);

				cmd_pipe_sync( rsp_dl );
				break;
			}

			case JumpToTranslated_CommandName_DList: {
				Command_JumpToTranslated const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_JumpToTranslated )
					);

				handle_lights_gpu(
					rsp_dl,
					self->context.lightsNum,
					self->context.luxes,
					getLast_Vec( &self->context.rspLights ),
					&self->context.rspAmbientLight,
					getLastConst_Vec( &self->context.cpuMtxs ),
					getLastConst_Vec( &self->context.materials )
				);
				if( self->context.cmd_SetRenderMode ) {
					cmd_set_render_mode_gpu(
						rsp_dl,
						Triangle_PrimitiveType_DList,
						self->context.zBufferEnabled,
						self->context.twoCycleMode,
						self->context.cmd_SetRenderMode
					);
				}
				SplitRequest_DLTrans request = cmd_jump_to_translated( rsp_dl, self->context.seg0_adr, microcode, cmd );
				if( true == request.request ) {
					addCommand_NativeDL( rsp_dl,
						gsDPFullSync()
					);
					addCommand_NativeDL( rsp_dl,
						gsSPEndDisplayList()
					);

					rewind_Stream( stream, sizeof *cmd );
					request.context = self->context;
					request.command = cmd;
					return request;
				}
				break;
			}

			case SetCombineMode_CommandName_DList: {
				Command_SetCombineMode const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetCombineMode )
					);

				self->context.cmd_SetCombineMode = cmd;
				cmd_set_combine_mode_gpu( rsp_dl, self->context.currentFragmentShader );
				break;
			}

			case SetRenderMode_CommandName_DList: {
				Command_SetRenderMode const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetRenderMode )
					);

				self->context.cmd_SetRenderMode = cmd;
				// execution deferred up to when reaching drawing commands
				// (DrawPrimitives / JumpToTranslated)
				break;
			}

			case NormalizePerspective_CommandName_DList: {
				Command_NormalizePerspective const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_NormalizePerspective )
					);

				self->context.cmd_NormalizePerspective = cmd;
				cmd_normalize_perspective( rsp_dl, cmd );
				break;
			}

			case SetLookAt_CommandName_DList: {
				abort();
#if 0
				Command_SetLookAt const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetLookAt )
					);

				cmd_set_look_at( rsp_dl, &self->context.current_lookat_i, cmd );
#endif
				break;
			}

			case SetGeometryMode_CommandName_DList: {
				Command_SetGeometryMode const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetGeometryMode )
					);

				self->context.cmd_SetGeometryMode = cmd;
				self->context.geometryMode =
					cmd_set_geometry_mode_gpu( rsp_dl, self->context.geometryMode, cmd->gmode );
				break;
			}

			case ClearGeometryMode_CommandName_DList: {
				Command_ClearGeometryMode const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_ClearGeometryMode )
					);

				self->context.geometryMode =
					cmd_clear_geometry_mode_gpu( rsp_dl, self->context.geometryMode, cmd->gmode );
				break;
			}

			case SetViewport_CommandName_DList: {
				Command_SetViewport const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetViewport )
					);

				self->context.cmd_SetViewport = cmd;
				cmd_set_viewport( rsp_dl, &self->context.rspVp, cmd );
				break;
			}

			case SetMaterial_CommandName_DList: {
				Command_SetMaterial const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetMaterial )
					);

				cmd_set_material(
					rsp_dl,
					&self->context.rspLights,
					&self->context.materials,
					&self->context.geometryMode,
					cmd
				);
				break;
			}

			case SetSegment_CommandName_DList: {
				Command_SetSegment const* cmd =
					extractElement_Stream( stream,
						sizeof( Command_SetSegment )
					);

				cmd_set_segment( &self->context, cmd );
				break;
			}

			case SetVertexShader_CommandName_DList: {
				cmd_set_vertex_shader(
					&self->context.currentVertexShader,
					(Command_SetVertexShader const*) extractElement_Stream( stream, sizeof( Command_SetVertexShader ) )
				);
				break;
			}

			case SetFragmentShader_CommandName_DList: {
				Command_SetFragmentShader const* cmd = extractElement_Stream( stream, sizeof( Command_SetFragmentShader ) );

				self->context.cmd_SetFragmentShader = cmd;
				cmd_set_fragment_shader( &self->context.currentFragmentShader, cmd );
				cmd_set_combine_mode_gpu( rsp_dl, self->context.currentFragmentShader );
				break;
			}

			case SetActiveTexturesMask_CommandName_DList: {
				cmd_set_active_textures_mask( rsp_dl, &self->context.activeTexturesMask, &self->context.twoCycleMode,
					(Command_SetActiveTexturesMask const*) extractElement_Stream( stream, sizeof( Command_SetActiveTexturesMask ) )
				);
				break;
			}

			// TODO
			//static bool prevZBufferEnabled;
			case PushState_CommandName_DList: {
				(Command_PushState const*)extractElement_Stream( stream, sizeof( Command_PushState ) );
				//prevZBufferEnabled = self->context.zBufferEnabled;
				self->context.zBufferEnabled = false;
				break;
			}

			// TODO
			case PopState_CommandName_DList: {
				(Command_PopState const*)extractElement_Stream( stream, sizeof( Command_PopState ) );
				self->context.zBufferEnabled = true; // prevZBufferEnabled;
				break;
			}
			
			default: {
				abort();
				break;
			}
		}
	} while( End_CommandName_DList != name );

	return (SplitRequest_DLTrans) {
		.request = false,
	};
}
