#include <assert.h>
#include <mempool.h>
#include <malloc.h>
#include <string.h>

static bool is_aligned( void const* adr, RegionAlignment const alignment ) {
	return !((uintptr_t)adr & ((1 << alignment) - 1));
}

static void* next_address( void const* current, size_t const el_size, RegionAlignment const alignment ) {
	uintptr_t next = (uintptr_t)current;
	do {
		next += el_size;
	} while( !is_aligned( (void const*)next, alignment ) );

	return (void*)next;
}

static void* align_address( void const* adr, RegionAlignment const alignment ) {
	return is_aligned( adr, alignment ) ? (void*)adr : next_address( adr, 1, alignment );
}

void init_MemPool( MemPool* self, size_t const el_size, size_t const num_el, RegionAlignment const alignment ) {
	assert( 0 == (num_el & (MEMPOOL_ALLOC_MAP_ENTRY_BITS - 1)) );

	self->adr_step = (uintptr_t)next_address( NULL, el_size, alignment );
	self->adr_step_scaled = MEMPOOL_ALLOC_MAP_ENTRY_BITS * self->adr_step;
	self->data = malloc( self->adr_step * num_el + (1 << (alignment - 1)) );
	assert( NULL != self->data );
	self->data_aligned = align_address( self->data, alignment );
	memset( self->alloc_map, 0, sizeof(self->alloc_map) );
}

void destroy_MemPool( MemPool* self ) {
	free( self->data );
}

void* allocate_MemPool( MemPool* self ) {
	uintptr_t entry_adr = (uintptr_t)self->data_aligned;

	for(
		uint_t i=0;
		i<MEMPOOL_ALLOC_MAP_ENTRIES;
		(++i), (entry_adr += self->adr_step_scaled)
	) {
		uint32_t bit_pattern = self->alloc_map[i];
		if( ~(uint32_t)0x0 == bit_pattern ) {
			continue;
		}

		uint32_t bit_test = 0x1;
		while( bit_pattern & bit_test ) {
			bit_test <<= 1;
			entry_adr += self->adr_step;
		}

		self->alloc_map[i] |= bit_test;

		return (void*)entry_adr;
	}

	abort();
	return NULL;
}

void return_MemPool( MemPool* self, void const* handle ) {
	uintptr_t const byteOffset = (uintptr_t)handle - (uintptr_t)self->data_aligned;
	uint_t const objectIndex = byteOffset / self->adr_step;
	uint_t const mapIndex = objectIndex / MEMPOOL_ALLOC_MAP_ENTRY_BITS;
	uint_t const bit = 1 << (objectIndex % MEMPOOL_ALLOC_MAP_ENTRY_BITS);

	assert( mapIndex < MEMPOOL_ALLOC_MAP_ENTRIES );
	assert( self->alloc_map[mapIndex] & bit );
	self->alloc_map[mapIndex] &= ~bit;
}
