#include <abort.h>
#include <sys/gfxdevice.h>
#include <ultra64.h>

#define BUFFER_W		320
#define BUFFER_H		(240/1)
#define	FRAMEBUFFER0	((void*)0x80200000)
#define	FRAMEBUFFER1	(void*)((uintptr_t)FRAMEBUFFER0 + BUFFER_W*BUFFER_H*2)
#define	AUXBUFFER		(void*)((uintptr_t)FRAMEBUFFER1 + BUFFER_W*BUFFER_H*2)
#define ZBUFFER			((void*)0x80300000)
////#define	ZBUFFER			(void*)((uintptr_t)AUXBUFFER + BUFFER_W*BUFFER_H*2)

static GfxMode const gfx_modes[] = {
	[0] = {
		.width = BUFFER_W,
		.height = BUFFER_H,
		.colorMode = GFXCOLORMODE_RGBA_5551
	}
};

static void waitForVBI_Sys( GfxDevice* self ) {
	if( MQ_IS_FULL( &self->viMessageQ ) ) {
		osRecvMesg(
			&self->viMessageQ,
			NULL,
			OS_MESG_BLOCK
		);
	}
	osRecvMesg(
		&self->viMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

void init_GfxDevice( GfxDevice* self ) {
	self->isRendering = false;
	self->frameCount = 0;

	{
		osCreateViManager( OS_PRIORITY_VIMGR );
		osViSetSpecialFeatures( OS_VI_GAMMA_DITHER_OFF );

		setMode_GfxDevice( self, 0 );

		osCreateMesgQueue( &self->viMessageQ, &self->viMessages, 1 );
		//osViSetEvent(&self->viMessageQ, NULL, 1);

		osCreateMesgQueue( &self->vbiMessageQ, &self->vbiMessages, 1 );
		osViSetEvent( &self->vbiMessageQ, NULL, 1 );
	}

	{
		osCreateMesgQueue( &self->rspMessageQ, &self->rspMessages, 1 );
		osSetEventMesg( OS_EVENT_SP, &self->rspMessageQ, NULL );
		
		osCreateMesgQueue( &self->rdpMessageQ, &self->rdpMessages, 1 );
		osSetEventMesg( OS_EVENT_DP, &self->rdpMessageQ, NULL );
	}

	{
		self->colorBuffer0 = create_Surface(
			Rgba5551_PixelFormat,
			FRAMEBUFFER0,
			BUFFER_W,
			BUFFER_H
		);
		self->colorBuffer1 = create_Surface(
			Rgba5551_PixelFormat,
			FRAMEBUFFER1,
			BUFFER_W,
			BUFFER_H
		);
		self->auxBuffer = create_Surface(
			Rgba5551_PixelFormat,
			AUXBUFFER,
			BUFFER_W,
			BUFFER_H
		);
		self->zBuffer = create_Surface(
			Int16_PixelFormat,
			ZBUFFER,
			BUFFER_W,
			BUFFER_H
		);
		self->colorBuffer = &self->colorBuffer0;
		osViSwapBuffer(
			getData_Surface( &self->colorBuffer1 )
		);
		waitForVBI_GfxDevice( self );
	}

	init_DLTrans( &self->dltrans );
	init_NativeDL( &self->displayList );

	setDefaultMicrocode_GfxDevice( self );
}

Surface const* getSurface_GfxDevice( GfxDevice const* self, SurfaceType_GfxDevice const type ) {
	switch( type ) {
		case Display_SurfaceType_GfxDevice:
			return self->colorBuffer == &self->colorBuffer0 ?
				&self->colorBuffer1 :
				&self->colorBuffer0;
		case Draw_SurfaceType_GfxDevice:
			return self->colorBuffer;
		case Z_SurfaceType_GfxDevice:
			return &self->zBuffer;
		case Aux_SurfaceType_GfxDevice:
			return &self->auxBuffer;
		default:
			abort();
	}
}

void swapBuffers_GfxDevice( GfxDevice* self ) {
	osViSwapBuffer(
		getData_Surface( self->colorBuffer )
	);

	waitForVBI_Sys( self );

	self->colorBuffer = (self->colorBuffer == &self->colorBuffer0) ?
		&self->colorBuffer1 : &self->colorBuffer0;
}


inline uint32_t getModeIndex_GfxDevice( GfxDevice const* self ) {
	return self->modeIdx;
}

inline GfxMode const* getCurrentMode_GfxDevice( GfxDevice const* self ) {
	return &gfx_modes[ self->modeIdx ];
}

void setMode_GfxDevice( GfxDevice* self, uint32_t const modeIdx ) {
	(0 == modeIdx) ? osViSetMode( &osViModePalLpn1 ) : abort();
	self->modeIdx = modeIdx;
}

GfxMode const* enumerateModes_GfxDevice( GfxDevice const* self ) {
	if( self ) {}

	return gfx_modes;
}

void waitForVBI_GfxDevice( GfxDevice const* self ) {
	GfxDevice* self_mut = (GfxDevice*)self;

	if( MQ_IS_FULL( &self->vbiMessageQ ) ) {
		osRecvMesg(
			&self_mut->vbiMessageQ,
			NULL,
			OS_MESG_BLOCK
		);
		self_mut->frameCount++;
	}
	osRecvMesg(
		&self_mut->vbiMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
	self_mut->frameCount++;

	/* forward notification to another queue */
	osJamMesg(
		&self_mut->viMessageQ,
		NULL,
		NULL
	);
}

static void waitForRSP( GfxDevice const* self ) {
	GfxDevice* self_mut = (GfxDevice*)self;
	
	osRecvMesg(
		&self_mut->rspMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

static void waitForRDP( GfxDevice const* self ) {
	GfxDevice* self_mut = (GfxDevice*)self;

	osRecvMesg(
		&self_mut->rdpMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

void waitForGPU_GfxDevice( GfxDevice const* self ) {
	if( false == isRendering_GfxDevice( self ) ) { return; }

	waitForRDP( self );
	waitForRSP( self );

	GfxDevice* self_mut = (GfxDevice*)self;
	self_mut->isRendering = false;
}

void runGPU_GfxDevice( GfxDevice const* self, DList const* dlist ) {
	GfxDevice* self_mut = (GfxDevice*)self;

	Stream stream = create_Stream( getData_DList( dlist ), getSize_DList( dlist ) );
	Context_DLTrans context;
	for( bool continueRequested = false; true; continueRequested = true ) {
		waitForGPU_GfxDevice( self_mut );
		resetCounter_NativeDL(
			getNativeDL_GfxDevice( self_mut )
		);

		SplitRequest_DLTrans splitRequest = process_DLTrans(
			&self_mut->dltrans,
			&stream,
			getMicrocodeConst_GfxDevice( self_mut )->id,
			dlist,
			getNativeDL_GfxDevice( self_mut ),
			continueRequested ? &context : NULL
		);

		{
			self_mut->tlist.t			= getMicrocodeConst_GfxDevice( self_mut )->tlist;
			self_mut->tlist.t.data_ptr	= (u64*)getHead_NativeDL( getNativeDL_GfxDevice( self_mut ) );
			self_mut->tlist.t.data_size	= getSizeInBytes_NativeDL( getNativeDL_GfxDevice( self_mut ) );
			self_mut->isRendering		= true;

			osWritebackDCacheAll();

			osSpTaskStart( &self_mut->tlist );
		}

		if( false == splitRequest.request ) {
			if( true == continueRequested ) {
				setDefaultMicrocode_GfxDevice( self_mut );
			}
			break;
		}

		context = splitRequest.context;

		{
			Microcode microcode;
			switch( splitRequest.microcodeRequested ) {
				case F3DEX2_MicrocodeId:
					microcode = create_Microcode( F3DEX2_MicrocodeId, (u64*)gspF3DEX2_xbusTextStart, (u64*)gspF3DEX2_xbusDataStart );
					break;
				case L3DEX2_MicrocodeId:
					microcode = create_Microcode( L3DEX2_MicrocodeId, (u64*)gspL3DEX2_xbusTextStart, (u64*)gspL3DEX2_xbusDataStart );
					break;
				default:	abort();
			};
			setMicrocode_GfxDevice( self_mut, microcode );
		}
	}
	destroy_Stream( &stream );
}

