#ifndef IMG_ACTION_H
#define IMG_ACTION_H

#include <engine.h>

typedef struct {
	GfxDevice const*	gfxDevice;
	Engine*				engine;
	Viewport const*		viewport;
	Texture const		*image0,
						*image1;
	Vector2u 			gridDivision;
	Vector4f			materialColor;
	Matrix*				transformation;
	float const*		opacity;
	void				(*shader_func)( Vertex_generic*, Vertex_generic const*, void const* );
} ImgActionArgs;

void* init_img_action( Time const* timeFromStart, Time const* timeDuration, void* arg );
void destroy_img_action( void* context, void* arg );
void handle_img_action( void* context, Time const* timeFromStart, void* arg );

void setImage_img_action( void* context, size_t const textureId );

Mesh createRectangleForTexture_img_action(
	Engine* engine,
	Vector2u const gridDivision_,
	Vector4f const* materialColor,
	Texture const* image0, Texture const* image1,
	GfxDevice const* gfxDevice
);

static inline Vector2f calculateScaleForImage_img_action( Texture const* texture ) {
	Surface const* surface = getSurface_GfxDevice( getGfxDevice_Sys(), Draw_SurfaceType_GfxDevice );
	float const screenAspectRatio = getAspectRatio_Surface( surface );
	return create_Vector2f(
		screenAspectRatio * getWidth_Texture( texture ) / getWidth_Surface( surface ),
		((float)getHeight_Texture( texture )) / getHeight_Surface( surface )
	);
}

#endif	/* IMG_ACTION_H */
