#include <modeler/modeler.h>
#include <sys/sys.h>
#include <task.h>
#include "img_action.h"
#include "main.h"

// TODO: turn off z-buffering

// problem: we don't own image0/1 - mutating it elswhere (width for example) invalidates mesh created
Mesh createRectangleForTexture_img_action(
	Engine* engine,
	Vector2u const gridDivision_,
	Vector4f const* materialColor,
	Texture const* image0, Texture const* image1,
	GfxDevice const* gfxDevice
) {
	if( gridDivision_.x ) {}
	assert( NULL != image0 );

	size_t const texelMemorySize = getTexelMemorySize_GfxDevice( gfxDevice,
		isColorIndexed_TextureFormat( getFormat_Texture( image0 ) )
	); //- (256 + 128 + 64);		// seam compensation
	//size_t const seamCompensation = getBytesPerPixel_Texture( image0 ) * 2 *		// *2, because it's 4bpp
	//	(getWidth_Texture( image0 ) + getHeight_Texture( image0 ) - 2) * 2;
	size_t const totalSize = getSizeOfPixels_Texture( image0 ); // + seamCompensation;
	size_t const numberOfTiles = (totalSize / texelMemorySize) +
		 ((totalSize % texelMemorySize) != 0);

	uint_t div = 1;
	//assert( numberOfTiles == 53 );
	for( ; (div * div) < numberOfTiles; ++div ) {}
	//if( div > 1 ) { ++div; }
	//while( log2_has_frac( div ) ) { ++div; }
	//div <<= 1;
	Vector2u const gridDivision = create_Vector2u( div, div );

	Mesh srcMesh = createPlane_Modeler( gridDivision.x, gridDivision.y );
	Layer const* srcLayer = getLayerConst_Mesh( &srcMesh, 0 );

	Mesh mesh = create_Mesh();

	size_t texture0Id, texture1Id;
	size_t const materialId =
		addMaterialReference_Engine( engine,
			createMaterial_Engine( engine,
				*materialColor,				//create_Vector4f( 1.0f, 1.0f, 1.0f, 1.0f ),
				(Ci4Ia16_TextureFormat == getFormat_Texture( image0 )) ?
					Alpha_TextureCombine : Decal_TextureCombine,
				(texture0Id = addTextureReference_Engine( engine, image0 )),
				(texture1Id = ((NULL == image1) ? TEXTURE_ID_NONE : addTextureReference_Engine( engine, image1 )))
			)
		);

	VecConstIterator triangles = createTriangleConstIterator_Layer( srcLayer );
	for( Triangle const* triangle; (triangle = next_VecConstIterator( &triangles )); ) {
		Layer layer = create_Layer( Triangle_PrimitiveType_Layer );
		setMaterialId_Layer( &layer, materialId );

		Vertex_generic v0, v1, v2;
		addVertex_Layer( &layer,
			v0 = *getVertexConst_Layer( srcLayer, triangle->ids[0] )
		);
		addVertex_Layer( &layer,
			v1 = *getVertexConst_Layer( srcLayer, triangle->ids[1] )
		);
		addVertex_Layer( &layer,
			v2 = *getVertexConst_Layer( srcLayer, triangle->ids[2] )
		);
		addTriangle_Layer( &layer, create_Triangle( 0, 1, 2 ) );

		Texture const* texture = getTextureConstReference_Engine( engine, texture0Id );
		size_t const width = getWidth_Texture( texture );
		size_t const height = getHeight_Texture( texture );
		size_t const sMin = (width - 1) * fminf(
			fminf( v0.texcoords.x, v1.texcoords.x ),
			v2.texcoords.x
		);
		size_t const tMin = (height - 1) * fminf(
			fminf( v0.texcoords.y, v1.texcoords.y ),
			v2.texcoords.y
		);
		size_t const sMax = (width - 1) * fmaxf(
			fmaxf( v0.texcoords.x, v1.texcoords.x ),
			v2.texcoords.x
		);
		size_t const tMax = (height - 1) * fmaxf(
			fmaxf( v0.texcoords.y, v1.texcoords.y ),
			v2.texcoords.y
		);
		layer.tile = (Rectangle) {				// inclusive
			.p0x = sMin, .p0y = tMin,
			.p1x = sMax, .p1y = tMax 			//+1?
		};
		addLayer_Mesh( &mesh, layer );
	}
	destroy_VecConstIterator( &triangles );

	destroy_Mesh( &srcMesh );

	return mesh;
}

typedef struct {
	Engine*			engine;
	Matrix*			transformation;
	float const*	opacity;
	float			t;
	VertexShader	shader0;
	Raw_Xformer		rootXf;
	Camera_Xformer	cameraXf;
	Mesh_Xformer	meshXf;
	Tree			sceneTree;
} Context;

static void init_cameraXf_Context( Camera_Xformer* cameraXf, Engine* engine ) {
	*cameraXf = createCamera_Engine( engine, Orthographic_CameraType );
	setParams_OrthoCamera( getOrtho_Camera( cameraXf->camera ),
		-1.0f, 1.0f,
		-1.0f, 1.0f,
		-1.0f, 1.0f
	);
}
static void destroy_cameraXf_Context( Camera_Xformer* cameraXf, Engine* engine ) {
	if( cameraXf ) {}
	if( engine ) {}
	// TODO
}

static void init_meshXf_Context(
	Mesh_Xformer* meshXf,
	Vector2u const gridDivision,
	Vector4f const* materialColor,
	Texture const* image0,
	Texture const* image1,
	VertexShader const* shader,
	Engine* engine,
	GfxDevice const* gfxDevice
) {
	*meshXf = createMesh_Engine( engine,
		createRectangleForTexture_img_action( engine, gridDivision, materialColor, image0, image1, gfxDevice ),
		NULL,
		shader
	);
}
static void destroy_meshXf_Context( Mesh_Xformer* meshXf, Engine* engine ) {
	freeMesh_Engine( engine, meshXf );
}

static void init_sceneTree_Context( Tree* sceneTree, Raw_Xformer* rootXf, Camera_Xformer* cameraXf, Mesh_Xformer* meshXf ) {

	init_Tree( sceneTree,
		create_TreeNode( rootXf, NULL,
			(TreeNode[]) {
				create_TreeNode( cameraXf, NULL, NULL, 0 ),
				create_TreeNode( meshXf, NULL, NULL, 0 )
			}, 2
		),
		NULL, NULL
	);
}
static void destroy_sceneTree_Context( Tree* sceneTree ) {
	destroy_Tree( sceneTree );
}


void* init_img_action( Time const* timeFromStart, Time const* timeDuration, void* arg ) {
	if( timeFromStart ) {}
	if( timeDuration ) {}

	Context* context = malloc( sizeof(*context) );

	ImgActionArgs const* args = arg;
	Engine* engine = args->engine;

	context->engine = engine;
	context->opacity = args->opacity;
	context->transformation = args->transformation;
	context->rootXf = createRaw_Engine( engine );
	init_cameraXf_Context( &context->cameraXf, args->engine );
	context->shader0 = create_VertexShader( args->shader_func, &context->t );
	init_meshXf_Context(
		&context->meshXf,
		args->gridDivision,
		&args->materialColor,
		args->image0,
		args->image1,
		args->shader_func ? &context->shader0 : NULL,
		args->engine,
		args->gfxDevice
	);
	init_sceneTree_Context( &context->sceneTree, &context->rootXf, &context->cameraXf, &context->meshXf );

	return context;
}

void destroy_img_action( void* context_generic, void* arg ) {
	if( arg ) {}

	Context* context = context_generic;

	destroy_sceneTree_Context( &context->sceneTree );
	destroy_meshXf_Context( &context->meshXf, context->engine );
	destroy_VertexShader( &context->shader0 );
	destroy_cameraXf_Context( &context->cameraXf, context->engine );
	freeRaw_Engine( context->engine, &context->rootXf );

	free( context );
}

void handle_img_action( void* context_generic, Time const* timeFromStart, void* arg ) {

	ImgActionArgs* args = arg;
	Context* context = context_generic;

	float timeNormalized = (float)asMilliseconds_Time( timeFromStart );

	if( NULL != context->opacity ) {
		setOpacity_Mesh( context->meshXf.mesh, *context->opacity );
	}

	if( NULL != context->transformation ) {
		transformLocally_Xformer( context->meshXf.xf, context->transformation );
	}

	Engine* engine = args->engine;
	Viewport const savedViewport = *getViewport_Engine( engine );
	Camera_Xformer const* savedCamera = getCamera_Engine( engine );
	Surface const* savedZSurface = getZBuffer_Engine( engine );

	setCamera_Engine( engine, &context->cameraXf );
	setViewport_Engine( engine, *args->viewport );
	setZBuffer_Engine( engine, NULL );

	context->t = timeNormalized / 1024.0f;
	process_task( engine, &(Task) {
		.f = (taskfptr)sceneTaskF_Engine,
		.arg = &context->sceneTree
	});

	setZBuffer_Engine( engine, savedZSurface );
	setViewport_Engine( engine, savedViewport );
	setCamera_Engine( engine, savedCamera );
}

void setImage_img_action( void* context_generic, size_t const textureId ) {
	Context* context = context_generic;

	Material* material =
		getMaterialReference_Engine( context->engine,
			getMaterialId_Layer( getFirstLayer_Mesh( context->meshXf.mesh ) )
		);
	setTextureId_Material( material, textureId );
}

