#include <stdlib/abort.h>
#include <actmgr/actmgr.h>
#include <containers/tree.h>
#include <engine.h>
#include <postproc/postproc.h>
#include <resmgr/resmgr.h>
#include <stdlib.h>
#include <sys/sys.h>
#include <sys/region.h>
#include <containers/vec_.h>
#include <string.h>
#include <malloc.h>
#include <mesh.h>
#include <textureGenerator.h>
#include "main.h"
#include "img_action.h"
#include "scene_action.h"
#include "clear_action.h"

static bool const
	sceneActionOn			= !false,
	imgActionOn				= !false,
	clearActionOn			= !false,
	tripModeOn				= false,
	postprocBlurOn			= false,
	audioOn					= !false;

typedef struct {
	Player	player;
	float 	angle;
} HandlerArg;

static void shader0_f( Vertex_generic* out, Vertex_generic const* in, void const* param ) {
	float const t = *(float const*)param;
	*out = *in;
	out->pos.x *= 1.01f;
	out->pos.y *= 1.01f;
	out->pos.x += 0.014f * (cosf( 4.0f * t + in->pos.x + in->pos.y ));
	out->pos.y += 0.018f * (sinf( 2.0f * t + 3.0f*in->pos.x + 2.0f*in->pos.y ));
}

static void handle_vbi( void* sys_arg, void* user_arg ) {
	if( sys_arg ) {}

	if( NULL != user_arg ) {
		*((float*)user_arg) += 1.0f;	// angle
	}
}

// TODO: Make some mutex to synchronize Player update with its reading by application
static void handle_controller( void* sys_arg, void* user_arg ) {
	static float stx_old,
				 sty_old,
				 ud_old;

	if( (NULL != sys_arg) && (NULL != user_arg) ) {
		OSContPad const* rdata = (OSContPad const*)sys_arg;
		Player* player = (Player*)user_arg;

		player->angle -= 0.04f * stx_old + rdata->stick_x * 0.015f;

		player->dir.x = cosf( M_PI * (player->angle + 90) / 180 );
		//player->dir_y = player->dir.y;
		player->dir.z = -sinf( M_PI * (player->angle + 90) / 180 );

		player->pos.x += 0.0165f * (rdata->stick_y + sty_old) * player->dir.x;
		player->pos.z += 0.0165f * (rdata->stick_y + sty_old) * player->dir.z;
		player->pos.y += 0.03f * ud_old + (rdata->button & U_JPAD && TRUE) - (rdata->button & D_JPAD && TRUE);

		sty_old = sty_old * 0.90f + 0.39f * rdata->stick_y;
		stx_old = 0.87f * stx_old + 0.012f * rdata->stick_x;
		ud_old = ud_old * 0.904f + (rdata->button & U_JPAD && TRUE) - (rdata->button & D_JPAD && TRUE);
	}
}

void process_task( Engine* engine, Task const* task ) {
	GfxDevice* gfxDevice = getGfxDevice_Sys();
	waitForGPU_GfxDevice( gfxDevice );

	processTask_Engine( engine, task );
	runGPU_GfxDevice( gfxDevice, getDisplayList_Engine( engine ) );
}

void mainproc( void* arg ) {
	if( arg ) {}

	if( true == audioOn ) {
		setSource_AudioDevice( getAudioDevice_Sys(),
			(void*)((uintptr_t)getResource_ResMgr( "sound" ) + 44)
		);
	}

	HandlerArg function_args = {
		.player = {
			.pos = (true == sceneActionOn) ?
				create_Vector3f( 0.0f, 130.0f, 200.0f ) :
				create_Vector3f( 0.0f, 0.0f, 30.0f ),
			.dir = create_Vector3f( 0.0f, 0.0f, 0.0f ),
			.angle = 0.0f
		},
		.angle = 0.0f
	};

	{
		registerHandler_Sys( Vbi_HandlerType_Sys, handle_vbi, &function_args.angle );
		registerHandler_Sys( Controller_HandlerType_Sys, handle_controller, &function_args );
	}

	Engine engine = create_Engine();
	ActMgr actmgr = create_ActMgr();

	#if 1
		ClearActionArgs clear_args = {
			.engine = &engine,
			.clearColor = create_Vector3f( 0.1f, 0.2f, 0.7f ),	//create_Vector3f( 0.3f, 0.7f, 0.8f ),
			.clearZ = 1.0f
		};
		Action clearAction = create_Action(
			init_clear_action,
			destroy_clear_action,
			handle_clear_action,
			&clear_args,
			(Time) { 0, 0, 0 },						// start
			(Time) ACTMGR_TIME_INFINITY				// duration
		);
		if( true == clearActionOn ) { addAction_ActMgr( &actmgr, clearAction ); }
	#endif

	#if 1
		SceneActionArgs scene_args = {
			.engine = &engine,
			.player = &function_args.player,
			.angle = &function_args.angle,
			.scalar_field_resolution = 9
		};
		Action sceneAction = create_Action(
			init_scene_action,
			destroy_scene_action,
			handle_scene_action,
			&scene_args,
			(Time) { 0, 0, 0 },						// start
			(Time) ACTMGR_TIME_INFINITY				// duration
		);
		if( true == sceneActionOn ) { addAction_ActMgr( &actmgr, sceneAction ); }
	#endif

	#if 1
		//Texture texture = ((StaticMesh*)getResource_ResMgr( "cpu" ))->textures[0];
			//generate_texture( 128, 128, Rgba16_TextureFormat, radial_func );
		Viewport imageViewport = create_Viewport( 0.25f, 0.45f, 0.75f, 0.55f, 0.0f, 1.0f );
		ImgActionArgs img_args = {
			.gfxDevice = getGfxDevice_Sys(),
			.engine = &engine,
			.viewport = &imageViewport,
			.image0 = &((StaticMesh*)getResource_ResMgr( "logo" ))->textures[0],
			.image1 = NULL,
			.gridDivision = create_Vector2u( 0, 0 ),	// ignored for now
			.materialColor = create_Vector4f( 1.0f, 1.0f, 1.0f, 1.0f ),
			.transformation = NULL,
			.opacity = NULL,
			.shader_func = shader0_f
		};
		Action imgAction = create_Action(
			init_img_action,
			destroy_img_action,
			handle_img_action,
			&img_args,
			(Time) { 0, 0, 0 },						// start
			(Time) ACTMGR_TIME_INFINITY				// duration
		);
		if( true == imgActionOn ) { addAction_ActMgr( &actmgr, imgAction ); }
	#endif
	
	GfxDevice* gfxDevice = getGfxDevice_Sys();
	setZBuffer_Engine( &engine, getSurface_GfxDevice( gfxDevice, Z_SurfaceType_GfxDevice ) );

	while( true ) {
		waitForGPU_GfxDevice( gfxDevice );
		swapBuffers_GfxDevice( gfxDevice );

		setRenderTarget_Engine( &engine,
			getSurface_GfxDevice( gfxDevice,
				( true == tripModeOn || true == postprocBlurOn) ?
					Aux_SurfaceType_GfxDevice :
					Draw_SurfaceType_GfxDevice
			)
		);

		Time const currentTime = getTime_Sys();
		process_ActMgr( &actmgr, &currentTime );

		if( true == postprocBlurOn ) {
			Surface const* src = getSurface_GfxDevice( gfxDevice, Aux_SurfaceType_GfxDevice );
			Surface const* dest = getSurface_GfxDevice( gfxDevice, Draw_SurfaceType_GfxDevice );
			postprocBlur( dest, src );
		}
	}
}

